"""
Test special KEYWORD rule.
"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest
from parglare import Parser, Grammar, RegExRecognizer, StringRecognizer
from parglare.exceptions import GrammarError, ParseError


def test_keyword_must_be_regex():
    grammar = r"""
    S: "for" name=ID "=" from=INT "to" to=INT;

    terminals
    KEYWORD: "id";
    ID: /\w+/;
    INT: /\d+/;
    """

    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'must have a regex recognizer defined' in str(e)


def test_keyword_grammar_init():
    grammar = r"""
    S: "for" name=ID "=" from=INT "to" to=INT;

    terminals
    KEYWORD: /\w+/;
    ID: /\w+/;
    INT: /\d+/;
    """

    g = Grammar.from_string(grammar)

    # 'for' term matches KEYWORD rule so it'll be replaced by
    # RegExRecognizer instance.
    for_term = g.get_terminal('for')
    assert type(for_term.recognizer) is RegExRecognizer
    assert for_term.recognizer._regex == r'\bfor\b'

    # '=' term doesn't match KEYWORD rule so it will not change
    eq_term = g.get_terminal('=')
    assert type(eq_term.recognizer) is StringRecognizer


def test_keyword_matches_on_word_boundary():
    grammar = r"""
    S: "for" name=ID "=" from=INT "to" to=INT EOF;

    terminals
    ID: /\w+/;
    INT: /\d+/;
    """

    g = Grammar.from_string(grammar)

    parser = Parser(g)
    # This will not raise an error
    parser.parse('forid=10 to20')

    # We add KEYWORD rule to the grammar to match ID-like keywords.
    grammar += r"KEYWORD: /\w+/;"

    g = Grammar.from_string(grammar)
    parser = Parser(g)
    with pytest.raises(ParseError) as e:
        # This *will* raise an error
        parser.parse('forid=10 to20')
    assert '"*forid=10 t" => Expected: for' in str(e)
    with pytest.raises(ParseError) as e:
        # This *will* also raise an error
        parser.parse('for id=10 to20')
    assert 'Expected: to' in str(e)

    # But this is OK
    parser.parse('for id=10 to 20')
    parser.parse('for for=10 to 20')


def test_keyword_preferred_over_regexes():
    """
    Test that keyword matches (internally converted to regex matches) are
    preferred over ordinary regex matches of the same length.
    """

    grammar = r"""
    S: "for"? name=ID? "=" from=INT "to" to=INT EOF;

    terminals
    ID: /\w+/;
    INT: /\d+/;
    KEYWORD: /\w+/;
    """
    g = Grammar.from_string(grammar)

    parser = Parser(g)

    # 'for' is ambiguous as it can be keyword or ID(name)
    # ParseError could be thrown but parglare will prefer
    # StringRecognizer and keywords over RegExRecognizer for
    # the match of the same length (i.e. "more specific match")
    parser.parse("for = 10 to 100")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
from os import path
import sys
import re
import itertools
from copy import copy
from parglare.six import add_metaclass
from parglare.exceptions import GrammarError, ParserInitError
from parglare.actions import pass_single, pass_none, collect, collect_sep
from parglare.common import Location, load_python_module
from parglare.termui import prints, s_emph, s_header, a_print, h_print
from parglare import termui

if sys.version < '3':
    text = unicode  # NOQA
else:
    text = str

# Associativity
ASSOC_NONE = 0
ASSOC_LEFT = 1
ASSOC_RIGHT = 2

# Priority
DEFAULT_PRIORITY = 10

# Multiplicity
MULT_ONE = '1'
MULT_OPTIONAL = '0..1'
MULT_ONE_OR_MORE = '1..*'
MULT_ZERO_OR_MORE = '0..*'

RESERVED_SYMBOL_NAMES = ['EOF', 'STOP', 'EMPTY']
SPECIAL_SYMBOL_NAMES = ['KEYWORD', 'LAYOUT']


def escape(instr):
    return instr.replace('\n', r'\n').replace('\t', r'\t')


class GrammarSymbol(object):
    """
    Represents an abstract grammar symbol.

    Attributes:
    name(str): The name of this grammar symbol.
    location(Location): The location where symbol is defined.
    action_name(string): Name of common/user action given in the grammar.
    action(callable): Resolved action given by the user. Overrides grammar
        action if provided. If not provided by the user defaults to
        grammar_action.
    grammar_action(callable): Resolved action given in the grammar.
    imported_with (PGFileImport): PGFileImport where this symbol is first time
        imported from. Used for FQN calculation.
    user_meta(dict): User meta-data.
    """
    def __init__(self, name, location=None, imported_with=None,
                 user_meta=None):
        self.name = escape(name)
        self.location = location
        self.action_name = None
        self.action = None
        self.grammar_action = None
        self.imported_with = imported_with
        self.user_meta = user_meta
        self._hash = hash(self.fqn)

    @property
    def fqn(self):
        if self.imported_with:
            return "{}.{}".format(self.imported_with.fqn, self.name)
        else:
            return self.name

    @property
    def action_fqn(self):
        if self.action_name:
            if self.imported_with:
                return "{}.{}".format(self.imported_with.fqn, self.action_name)
            else:
                return self.action_name

    def add_user_meta_data(self, name, value):
        if self.user_meta is None:
            self.user_meta = {}
        self.user_meta[name] = value

    def __getattr__(self, name):
        if self.user_meta is not None:
            attr = self.user_meta.get(name)
            if attr:
                return attr
        raise AttributeError

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return self.fqn

    def __repr__(self):
        return "{}({})".format(type(self).__name__, str(self))

    def __hash__(self):
        return self._hash


class NonTerminal(GrammarSymbol):
    """Represents a non-termial symbol of the grammar.

    Attributes:
    productions(list of Production): A list of alternative productions for
        this NonTerminal.
    """
    def __init__(self, name, productions=None, location=None,
                 imported_with=None, user_meta=None):
        super(NonTerminal, self).__init__(name, location, imported_with,
                                          user_meta)
        self.productions = productions if productions is not None else []


class Terminal(GrammarSymbol):
    """Represent a terminal symbol of the grammar.

    Attributes:
    prior(int): Priority used for lexical disambiguation.
    dynamic(bool): Should dynamic disambiguation be called to resolve conflict
        involving this terminal.
    finish(bool): Used for scanning optimization. If this terminal is `finish`
        no other recognizers will be checked if this succeeds. If not provided
        in the grammar implicit rules will be used during table construction.
    prefer(bool): Prefer this recognizer in case of multiple recognizers match
        at the same place and implicit disambiguation doesn't resolve.
    keyword(bool): `True` if this Terminal represents keyword. `False` by
        default.

    recognizer(callable): Called with input list of objects and position in the
        stream. Should return a sublist of recognized objects. The sublist
        should be rooted at the given position.
    """
    def __init__(self, name, recognizer=None, location=None,
                 imported_with=None):
        self.prior = DEFAULT_PRIORITY
        self._recognizer = None
        self.recognizer = recognizer if recognizer else StringRecognizer(name)
        self.finish = None
        self.prefer = False
        self.dynamic = False
        self.keyword = False
        super(Terminal, self).__init__(name, location, imported_with,
                                       user_meta=None)

    @property
    def recognizer(self):
        return self._recognizer

    @recognizer.setter
    def recognizer(self, value):
        self._recognizer = value


class Reference(object):
    """
    A name reference to a GrammarSymbol used for cross-resolving during
    grammar construction.
    Attributes:
        name (str): The FQN name of the referred symbol. This is the name of
            the original desuggared symbol without taking into account
            multiplicity and separator.
        location (Location): Location object of this reference.
        multiplicty(str): Multiplicity of the RHS reference (used for regex
            operators ?, *, +). See MULT_* constants above. By default
            multiplicity is MULT_ONE.
        separator (symbol or Reference): A reference to the separator symbol or
            the separator symbol itself if resolved.
    """
    def __init__(self, location, name):
        self.name = name
        self.location = location
        self.multiplicity = MULT_ONE
        self.separator = None

    @property
    def multiplicity_name(self):
        """
        Returns the name of the symbol that should be used if
        multiplicity/separator is used.
        """
        return make_multiplicity_name(
            self.name, self.multiplicity,
            self.separator.name if self.separator else None)

    def clone(self):
        new_ref = Reference(self.location, self.name)
        new_ref.multiplicity = self.multiplicity
        new_ref.separator = self.separator
        return new_ref

    def __repr__(self):
        return self.name


class Recognizer(object):
    """
    Recognizers are callables capable of recognizing low-level patterns
    (a.k.a tokens) in the input.
    """
    def __init__(self, name, location=None):
        self.name = name
        self.location = location


class StringRecognizer(Recognizer):
    def __init__(self, value, ignore_case=False, **kwargs):
        super(StringRecognizer, self).__init__(value, **kwargs)
        self.value = value
        self.ignore_case = ignore_case
        self.value_cmp = value.lower() if ignore_case else value

    def __call__(self, in_str, pos):
        if self.ignore_case:
            if in_str[pos:pos+len(self.value)].lower() == self.value_cmp:
                return self.value
        else:
            if in_str[pos:pos+len(self.value)] == self.value_cmp:
                return self.value


def esc_control_characters(regex):
    """
    Escape control characters in regular expressions.
    """
    unescapes = [('\a', r'\a'), ('\b', r'\b'), ('\f', r'\f'), ('\n', r'\n'),
                 ('\r', r'\r'), ('\t', r'\t'), ('\v', r'\v')]
    for val, text in unescapes:
        regex = regex.replace(val, text)
    return regex


class RegExRecognizer(Recognizer):
    def __init__(self, regex, name=None, re_flags=re.MULTILINE,
                 ignore_case=False, **kwargs):
        if name is None:
            name = regex
        super(RegExRecognizer, self).__init__(name, kwargs)
        self._regex = regex
        self.ignore_case = ignore_case
        if ignore_case:
            re_flags |= re.IGNORECASE
        self.re_flags = re_flags
        try:
            self.regex = re.compile(self._regex, re_flags)
        except re.error as ex:
            regex = esc_control_characters(self._regex)
            message = 'Regex compile error in /{}/ (report: "{}")'
            raise GrammarError(message.format(regex, str(ex)))

    def __call__(self, in_str, pos):
        m = self.regex.match(in_str, pos)
        if m and m.group():
            return m.group()


def EMPTY_recognizer(input, pos):
    pass


def EOF_recognizer(input, pos):
    pass


def STOP_recognizer(input, pos):
    pass


# These two terminals are special terminals used internally.
AUGSYMBOL = NonTerminal("S'")
STOP = Terminal("STOP", STOP_recognizer)

# These two terminals are special terminals used in the grammars.
# EMPTY will match nothing and always succeed.
# EOF will match only at the end of the input string.
EMPTY = Terminal("EMPTY", EMPTY_recognizer)
EMPTY.grammar_action = pass_none
EOF = Terminal("EOF", EOF_recognizer)
EOF.grammar_action = pass_none


class Production(object):
    """Represent production from the grammar.

    Attributes:
    symbol (GrammarSymbol):
    rhs (ProductionRHS):
    assignments(dict): Assignment instances keyed by name.
    assoc (int): Associativity. Used for ambiguity (shift/reduce) resolution.
    prior (int): Priority. Used for ambiguity (shift/reduce) resolution.
    dynamic (bool): Is dynamic disambiguation used for this production.
    nops (bool): Disable prefer_shifts strategy for this production.
        Only makes sense for GLR parser.
    nopse (bool): Disable prefer_shifts_over_empty strategy for this
        production. Only makes sense for GLR parser.
    user_meta(dict): User meta-data.
    prod_id (int): Ordinal number of the production.
    prod_symbol_id (int): A zero-based ordinal of alternative choice for this
        production grammar symbol.
    """

    def __init__(self, symbol, rhs, assignments=None, assoc=ASSOC_NONE,
                 prior=DEFAULT_PRIORITY, dynamic=False, nops=False,
                 nopse=False, user_meta=None):
        """
        Args:
        symbol (GrammarSymbol): A grammar symbol on the LHS of the production.
        rhs (list of GrammarSymbols):
        """
        self.symbol = symbol
        self.rhs = rhs if rhs else ProductionRHS()
        self.assignments = None
        if assignments:
            self.assignments = {}
            for assignment in assignments:
                if assignment.name:
                    self.assignments[assignment.name] = assignment
        self.assoc = assoc
        self.prior = prior
        self.dynamic = dynamic
        self.nops = nops
        self.nopse = nopse
        self.user_meta = user_meta

    def __str__(self):
        if hasattr(self, 'prod_id'):
            return (s_header("%d:") + " %s " + s_emph("=") +
                    " %s") % (self.prod_id, self.symbol, self.rhs)
        else:
            return ("%s " + s_emph("=") + " %s") % (self.symbol, self.rhs)

    def __repr__(self):
        return 'Production({})'.format(str(self.symbol))

    def __getattr__(self, name):
        if self.user_meta is not None:
            attr = self.user_meta.get(name)
            if attr:
                return attr
        raise AttributeError


class ProductionRHS(list):
    def __getitem__(self, idx):
        try:
            while True:
                symbol = super(ProductionRHS, self).__getitem__(idx)
                if symbol is not EMPTY:
                    break
                idx += 1
            return symbol
        except IndexError:
            return None

    def __len__(self):
        return super(ProductionRHS, self).__len__() - self.count(EMPTY)

    def __str__(self):
        return " ".join([str(x) for x in self])

    def __repr__(self):
        return "<ProductionRHS([{}])>".format(
            ", ".join([str(x) for x in self]))


class Assignment(object):
    """
    General assignment (`=` or `?=`, a.k.a. `named matches`) in productions.
    Used also for references as LHS and assignment operator are optional.
    """
    def __init__(self, name, op, symbol):
        """
        Attributes:
            name(str): The name on the LHS of assignment.
            op(str): Either a `=` or `?=`.
            symbol(Reference or GrammarSymbol): A grammar symbol on the RHS.
            symbol_name(str): A de-sugarred grammar symbol name on the
                RHS, i.e. referenced symbol without regex operators.
            multiplicty(str): Multiplicity of the RHS reference (used for regex
                operators ?, *, +). See MULT_* constants above. By default
                multiplicity is MULT_ONE.
            index(int): Index in the production RHS
        """
        self.name = name
        self.op = op
        self.symbol = symbol
        self.symbol_name = symbol.name
        self.multiplicity = symbol.multiplicity \
            if isinstance(symbol, Reference) else MULT_ONE
        self.index = None


class PGAttribute(object):
    """
    PGAttribute definition created by named matches.

    Attributes:
        name(str): The name of the attribute.
        multiplicity(str): Multiplicity of the attribute. See MULT_* constants.
        type_name(str): The type name of the attribute value(s). It is also the
            name of the referring grammar rule.
    """
    def __init__(self, name, multiplicity, type_name):
        self.name = name
        self.multiplicity = multiplicity
        self.type_name = type_name


class PGFile(object):
    """Objects of this class represent parglare grammar files.

    Grammar files can be imported using `import` keyword. Rules referenced from
    the imported grammar must be fully qualified by the grammar module name. By
    default the name of the target .pg file is the name of the module. `as`
    keyword can be used to override the default.

    Example:
    ```
    import `some/path/mygrammar.pg` as target
    ```

    Rules from file `mygrammar.pg` will be available under `target` namespace:

    ```
    MyRule: target.someRule+;
    ```

    Actions are by default loaded from the file named `<grammar>_actions.py`
    where `grammar` is basename of grammar file. Recognizers are loaded from
    `<grammar>_recognizers.py`. Actions and recognizers given this way are both
    optional. Furthermore, both actions and recognizers can be overriden by
    supplying actions and/or recognizers dict during grammar/parser
    instantiation.

    Attributes:

    productions (list of Production): Local productions defined in this file.
    terminals (dict of Terminal):
    classes (dict of ParglareClass): Dynamically created classes. Used by
        obj action.
    imports (dict): Mapping imported module/file local name to PGFile object.
    file_path (str): A full canonic path to the .pg file.
    grammar (PGFile): A root/grammar file.
    recognizers (dict of callables): A dict of Python callables used as a
        terminal recognizers.
    """
    def __init__(self, productions, terminals=None, classes=None, imports=None,
                 file_path=None, grammar=None, recognizers=None,
                 imported_with=None):
        self.productions = productions
        self.terminals = terminals
        self.classes = classes if classes else {}
        self.grammar = self if grammar is None else grammar

        self.file_path = path.realpath(file_path) if file_path else None
        self.imported_with = imported_with
        self.recognizers = recognizers
        self.actions = {}

        self.collect_and_unify_symbols()

        if self.file_path:
            self.grammar.imported_files[self.file_path] = self

        if imports:
            self.imports = {i.module_name: i for i in imports}
            for i in imports:
                i.grammar = self.grammar
                try:
                    i.load_pgfile()
                except IOError:
                    raise GrammarError(
                        location=Location(file_name=self.file_path),
                        message='Can\'t import file "{}".'.format(
                            i.file_path))
        else:
            self.imports = {}

        self.resolve_references()
        self.load_actions()
        self.load_recognizers()

    def collect_and_unify_symbols(self):
        """Collect non-terminals and terminals (both explicit and implicit/inline)
        defined in this file and make sure there is only one instance for each
        of them.

        """
        nonterminals_by_name = {}
        terminals_by_name = {}
        terminals_by_str_rec = {}

        # Check terminal uniqueness in both name and string recognition
        # and collect all terminals from explicit definitions.
        for terminal in self.terminals:
            if terminal.name in terminals_by_name:
                raise GrammarError(
                    location=terminal.location,
                    message='Multiple definitions of terminal rule "{}"'
                            .format(terminal.name))
            if isinstance(terminal.recognizer, StringRecognizer):
                rec = terminal.recognizer
                if rec.value in terminals_by_str_rec:
                    raise GrammarError(
                        location=terminal.location,
                        message='Terminals "{}" and "{}" match '
                        'the same string.'
                        .format(terminal.name,
                                terminals_by_str_rec[rec.value].name))
                terminals_by_str_rec[rec.value] = terminal
            terminals_by_name[terminal.name] = terminal

        self.terminals = terminals_by_name
        self.terminals_by_str_rec = terminals_by_str_rec

        # Collect non-terminals
        for production in self.productions:
            symbol = production.symbol
            symbol.imported_with = self.imported_with
            # Check that there is no terminal defined by the same name.
            if symbol.name in self.terminals:
                raise GrammarError(
                    location=symbol.location,
                    message='Rule "{}" already defined as terminal'
                    .format(symbol.name))
            # Unify all non-terminal objects
            if symbol.name in nonterminals_by_name:
                old_symbol = symbol
                new_symbol = nonterminals_by_name[symbol.name]
                production.symbol = new_symbol
            else:
                nonterminals_by_name[symbol.name] = symbol
                old_symbol = new_symbol = symbol
            new_symbol.productions.append(production)

            # Check grammar actions for rules/symbols.
            if new_symbol.action_name:
                if new_symbol.action_name != old_symbol.action_name:
                    raise GrammarError(
                        location=new_symbol.location,
                        message='Multiple different grammar actions '
                        'for rule "{}".'.format(new_symbol.name))

        self.nonterminals = nonterminals_by_name
        self.symbols_by_name = dict(nonterminals_by_name)
        self.symbols_by_name.update(self.terminals)

        # Add special terminals
        self.symbols_by_name['EMPTY'] = EMPTY
        self.symbols_by_name['EOF'] = EOF
        self.symbols_by_name['STOP'] = STOP

    def resolve_references(self):
        # Two pass resolving to enable referening symbols created during
        # resolving (e.g. multiplicity symbols).
        for pazz in [True, False]:
            for production in self.productions:
                for idx, ref in enumerate(production.rhs):
                    if isinstance(ref, Reference):
                        production.rhs[idx] = self.resolve_ref(ref, pazz)

    def register_symbol(self, symbol):
        self.symbols_by_name[symbol.name] = symbol

    def load_actions(self):
        """
        Loads actions from <grammar_name>_actions.py if the file exists.
        Actions must be collected with action decorator and the decorator must
        be called `action`.
        """
        actions_file = None
        if self.file_path:
            actions_file = path.join(
                path.dirname(self.file_path),
                "{}_actions.py".format(path.splitext(
                    path.basename(self.file_path))[0]))
            if path.exists(actions_file):
                mod_name = "{}actions".format(
                    self.imported_with.fqn
                    if self.imported_with is not None else "")
                actions_module = load_python_module(mod_name, actions_file)
                if not hasattr(actions_module, 'action'):
                    raise GrammarError(
                        Location(file_name=actions_file),
                        message='Actions file "{}" must have "action" '
                        'decorator defined.'.format(actions_file))
                self.actions = actions_module.action.all

    def load_recognizers(self):
        """Load recognizers from <grammar_name>_recognizers.py. Override
        with provided recognizers.

        """
        if self.file_path:
            recognizers_file = path.join(
                path.dirname(self.file_path),
                "{}_recognizers.py".format(path.splitext(
                    path.basename(self.file_path))[0]))

            if path.exists(recognizers_file):
                mod_name = "{}recognizers".format(
                    self.imported_with.fqn
                    if self.imported_with is not None else "")
                mod_recognizers = load_python_module(mod_name,
                                                     recognizers_file)
                recognizers = mod_recognizers.recognizer.all

                for recognizer_name, recognizer in recognizers.items():
                    symbol = self.resolve_symbol_by_name(
                        recognizer_name,
                        location=Location(file_name=recognizers_file))
                    if symbol is None:
                        raise GrammarError(
                            location=Location(file_name=recognizers_file),
                            message='Recognizer given for unknown '
                            'terminal "{}".'.format(recognizer_name)
                        )
                    if not isinstance(symbol, Terminal):
                        raise GrammarError(
                            location=Location(file_name=recognizers_file),
                            message='Recognizer given for non-terminal "{}".'
                            .format(recognizer_name))
                    symbol.recognizer = recognizer

    def resolve_ref(self, symbol_ref, first_pass=False):
        """Resolves given symbol reference.

        For local name search this file, for FQN use imports and delegate to
        imported file.

        On each resolved symbol productions in the root file are updated.

        If this is first pass do not fail on unexisting reference as there
        might be new symbols created during resolving (e.g. multiplicity
        symbols).

        """
        if isinstance(symbol_ref.separator, Reference):
            symbol_ref.separator = self.resolve_ref(symbol_ref.separator)

        symbol_name = symbol_ref.name
        symbol = self.resolve_symbol_by_name(symbol_name, symbol_ref.location)
        if not symbol:
            if first_pass:
                return symbol_ref
            else:
                raise GrammarError(
                    location=symbol_ref.location,
                    message='Unknown symbol "{}"'.format(symbol_name))

        mult = symbol_ref.multiplicity
        if mult != MULT_ONE:
            # If multiplicity is used than we are referring to
            # suggared symbol

            separator = symbol_ref.separator \
                if symbol_ref.separator else None

            base_symbol = symbol
            symbol_name = symbol_ref.multiplicity_name
            symbol = self.resolve_symbol_by_name(symbol_name,
                                                 symbol_ref.location)
            if not symbol:
                # If there is no multiplicity version of the symbol we
                # will create one at this place
                symbol = self.make_multiplicity_symbol(
                    symbol_ref, base_symbol, separator, self.imported_with)

        return symbol

    def resolve_symbol_by_name(self, symbol_name, location=None):
        """
        Resolves symbol by fqn.
        """
        if '.' in symbol_name:
            import_module_name, name = symbol_name.split('.', 1)
            try:
                imported_pg_file = self.imports[import_module_name]
            except KeyError:
                raise GrammarError(
                    location=location,
                    message='Unexisting module "{}" in reference "{}"'
                    .format(import_module_name, symbol_name))
            return imported_pg_file.resolve_symbol_by_name(name, location)
        else:
            return self.symbols_by_name.get(symbol_name, None)

    def resolve_action_by_name(self, action_name):
        if action_name in self.actions:
            return self.actions[action_name]
        elif '.' in action_name:
            import_module_name, name = action_name.split('.', 1)
            if import_module_name in self.imports:
                imported_pg_file = self.imports[import_module_name]
                return imported_pg_file.resolve_action_by_name(name)

    def make_multiplicity_symbol(self, symbol_ref, base_symbol, separator,
                                 imported_with):
        """
        Creates new NonTerminal for symbol refs using multiplicity and
        separators.
        """
        mult = symbol_ref.multiplicity
        if mult in [MULT_ONE_OR_MORE, MULT_ZERO_OR_MORE]:
            symbol_name = make_multiplicity_name(
                symbol_ref.name, MULT_ONE_OR_MORE,
                separator.name if separator else None)
            symbol = self.symbols_by_name.get(symbol_name)
            if not symbol:
                # noqa See: http://www.igordejanovic.net/parglare/grammar_language/#one-or-more_1
                productions = []
                symbol = NonTerminal(symbol_name, productions,
                                     base_symbol.location,
                                     imported_with=imported_with)

                if separator:
                    productions.append(
                        Production(symbol,
                                   ProductionRHS([symbol,
                                                  separator,
                                                  base_symbol])))
                    symbol.action_name = 'collect_sep'
                else:
                    productions.append(
                        Production(symbol,
                                   ProductionRHS([symbol,
                                                  base_symbol])))
                    symbol.action_name = 'collect'

                productions.append(
                    Production(symbol, ProductionRHS([base_symbol])))

                self.register_symbol(symbol)

            if mult == MULT_ZERO_OR_MORE:
                productions = []
                symbol_one = symbol
                symbol_name = make_multiplicity_name(
                    symbol_ref.name, mult,
                    separator.name if separator else None)
                symbol = NonTerminal(symbol_name, productions,
                                     base_symbol.location,
                                     imported_with=imported_with)

                productions.extend([Production(symbol,
                                               ProductionRHS([symbol_one]),
                                               nops=True),
                                    Production(symbol,
                                               ProductionRHS([EMPTY]))])

                def action(_, nodes):
                    if nodes:
                        return nodes[0]
                    else:
                        return []

                symbol.grammar_action = action

                self.register_symbol(symbol)

        else:
            # MULT_OPTIONAL
            if separator:
                raise GrammarError(
                    location=symbol_ref.location,
                    message='Repetition modifier not allowed for '
                    'optional (?) for symbol "{}".'
                    .format(symbol_ref.name))
            productions = []
            symbol_name = make_multiplicity_name(symbol_ref.name, mult)
            symbol = NonTerminal(symbol_name, productions,
                                 base_symbol.location,
                                 imported_with=imported_with)
            productions.extend([Production(symbol,
                                           ProductionRHS([base_symbol])),
                                Production(symbol,
                                           ProductionRHS([EMPTY]))])

            symbol.action_name = 'optional'

            self.register_symbol(symbol)

        return symbol


class Grammar(PGFile):
    """
    Grammar is a collection of production rules, nonterminals and terminals.
    First production is reserved for the augmented production (S' -> S).

    Attributes:
    start_symbol (GrammarSymbol or str): start/root symbol of the grammar or
        its name.
    nonterminals (set of NonTerminal):
    terminals(set of Terminal):
    imported_files(dict): Global registry of all imported files.

    """

    def __init__(self, productions=None, terminals=None,
                 classes=None, imports=None, file_path=None, recognizers=None,
                 start_symbol=None, _no_check_recognizers=False,
                 re_flags=re.MULTILINE, ignore_case=False, debug=False,
                 debug_parse=False, debug_colors=False):
        """
        Grammar constructor is not meant to be called directly by the user.
        See `from_str` and `from_file` static methods instead.

        Arguments:
        see Grammar attributes.
        _no_check_recognizers (bool, internal): Used by pglr tool to circumvent
             errors for empty recognizers that will be provided in user code.
        """

        self.imported_files = {}

        super(Grammar, self).__init__(productions=productions,
                                      terminals=terminals,
                                      classes=classes,
                                      imports=imports,
                                      file_path=file_path,
                                      grammar=self,
                                      recognizers=recognizers)

        self._no_check_recognizers = _no_check_recognizers

        # Determine start symbol. If name is provided search for it. If name is
        # not given use the first production LHS symbol as the start symbol.
        if start_symbol:
            if isinstance(start_symbol, str):
                for p in self.productions:
                    if p.symbol.name == start_symbol:
                        self.start_symbol = p.symbol
            else:
                self.start_symbol = start_symbol
        else:
            # By default, first production symbol is the start symbol.
            self.start_symbol = self.productions[0].symbol

        self._init_grammar()

    def _init_grammar(self):
        """
        Extracts all grammar symbol (nonterminal and terminal) from the
        grammar, resolves and check references in productions, unify all
        grammar symbol objects and enumerate productions.
        """
        # Reserve 0 production. It is used for augmented prod. in LR
        # automata calculation.
        self.productions.insert(
            0,
            Production(AUGSYMBOL, ProductionRHS([self.start_symbol, STOP])))

        self._add_all_symbols_productions()
        self._enumerate_productions()
        self._fix_keyword_terminals()
        self._resolve_actions()

        # Connect recognizers, override grammar provided
        if not self._no_check_recognizers:
            self._connect_override_recognizers()

    def _add_all_symbols_productions(self):

        self.nonterminals = {}
        for prod in self.productions:
            self.nonterminals[prod.symbol.fqn] = prod.symbol
        self.terminals.update([(s.name, s) for s in (EMPTY, EOF, STOP)])

        def add_productions(productions):
            for production in productions:
                symbol = production.symbol
                if symbol.fqn not in self.nonterminals:
                    self.nonterminals[symbol.fqn] = symbol
                for idx, rhs_elem in enumerate(production.rhs):
                    if isinstance(rhs_elem, Terminal):
                        if rhs_elem.fqn not in self.terminals:
                            self.terminals[rhs_elem.fqn] = rhs_elem
                        else:
                            # Unify terminals
                            production.rhs[idx] = self.terminals[rhs_elem.fqn]
                    elif isinstance(rhs_elem, NonTerminal):
                        if rhs_elem.fqn not in self.nonterminals:
                            self.productions.extend(rhs_elem.productions)
                            add_productions(rhs_elem.productions)
                    else:
                        # This should never happen
                        assert False, "Invalid RHS element type '{}'."\
                            .format(type(rhs_elem))
        add_productions(list(self.productions))

    def _enumerate_productions(self):
        """
        Enumerates all productions (prod_id) and production per symbol
        (prod_symbol_id).
        """
        idx_per_symbol = {}
        for idx, prod in enumerate(self.productions):
            prod.prod_id = idx
            prod.prod_symbol_id = idx_per_symbol.get(prod.symbol, 0)
            idx_per_symbol[prod.symbol] = \
                idx_per_symbol.get(prod.symbol, 0) + 1

    def _fix_keyword_terminals(self):
        """
        If KEYWORD terminal with regex match is given fix all matching string
        recognizers to match on a word boundary.
        """
        keyword_term = self.get_terminal('KEYWORD')
        if keyword_term is None:
            return

        # KEYWORD rule must have a regex recognizer
        keyword_rec = keyword_term.recognizer
        if not isinstance(keyword_rec, RegExRecognizer):
            raise GrammarError(
                location=keyword_term.location,
                message='KEYWORD rule must have a regex recognizer defined.')

        # Change each string recognizer corresponding to the KEYWORD
        # regex by the regex recognizer that match on word boundaries.
        for term in self.terminals.values():
            if isinstance(term.recognizer, StringRecognizer):
                match = keyword_rec(term.recognizer.value, 0)
                if match == term.recognizer.value:
                    term.recognizer = RegExRecognizer(
                        r'\b{}\b'.format(match),
                        ignore_case=term.recognizer.ignore_case)
                    term.keyword = True

    def _resolve_actions(self, action_overrides=None,
                         fail_on_no_resolve=False):
        """
        Checks and resolves semantic actions given in the grammar and
        additional `*_actions.py` module.

        Args:
            action_overrides(dict): Dict of actions that take precendence. Used
                for actions supplied during parser construction.
        """
        import parglare.actions as actmodule

        for symbol in self:

            # Resolve trying from most specific to least specific
            action = None

            # 1. Resolve by fully qualified symbol name
            if '.' in symbol.fqn:
                if action_overrides:
                    action = action_overrides.get(symbol.fqn, None)

                if action is None:
                    action = self.resolve_action_by_name(symbol.fqn)

            # 2. Fully qualified action name
            if action is None and symbol.action_fqn is not None \
               and '.' in symbol.action_fqn:
                if action_overrides:
                    action = action_overrides.get(symbol.action_fqn, None)

                if action is None:
                    action = self.resolve_action_by_name(symbol.action_fqn)

            # 3. Symbol name
            if action is None:
                if action_overrides:
                    action = action_overrides.get(symbol.name, None)

                if action is None:
                    action = self.resolve_action_by_name(symbol.name)

            # 4. Action name
            if action is None and symbol.action_name is not None:
                if action_overrides:
                    action = action_overrides.get(symbol.action_name, None)

                if action is None:
                    action = self.resolve_action_by_name(symbol.action_name)

                # 5. Try to find action in built-in actions module.
                if action is None:
                    action_name = symbol.action_name
                    if hasattr(actmodule, action_name):
                        action = getattr(actmodule, action_name)

            if symbol.action_name and action is None \
               and fail_on_no_resolve:
                raise ParserInitError(
                    'Action "{}" given for rule "{}" '
                    'doesn\'t exists in parglare common actions and '
                    'is not provided using "actions" parameter.'
                    .format(symbol.action_name, symbol.name))

            if action is not None:
                symbol.action = action

                # Some sanity checks for actions
                if type(symbol.action) is list:
                    if type(symbol) is Terminal:
                        raise ParserInitError(
                            'Cannot use a list of actions for '
                            'terminal "{}".'.format(symbol.name))
                    else:
                        if len(symbol.action) != len(symbol.productions):
                            raise ParserInitError(
                                'Length of list of actions must match the '
                                'number of productions for non-terminal '
                                '"{}".'.format(symbol.name))
            else:
                symbol.action = symbol.grammar_action

    def _connect_override_recognizers(self):
        for term in self.terminals.values():
            if self.recognizers and term.fqn in self.recognizers:
                term.recognizer = self.recognizers[term.fqn]
            else:
                if term.recognizer is None:
                    if not self.recognizers:
                        raise GrammarError(
                            location=term.location,
                            message='Terminal "{}" has no recognizer defined '
                            'and no recognizers are given during grammar '
                            'construction.'.format(term.fqn))
                    else:
                        if term.fqn not in self.recognizers:
                            raise GrammarError(
                                location=term.location,
                                message='Terminal "{}" has no recognizer '
                                'defined.'.format(term.fqn))

    def get_terminal(self, name):
        "Returns terminal with the given fully qualified name or name."
        return self.terminals.get(name)

    def get_nonterminal(self, name):
        "Returns non-terminal with the given fully qualified name or name."
        return self.nonterminals.get(name)

    def get_symbol(self, name):
        "Returns grammar symbol with the given name."
        s = self.get_terminal(name)
        if not s:
            s = self.get_nonterminal(name)
        return s

    def __iter__(self):
        return (s for s in itertools.chain(self.nonterminals.values(),
                                           self.terminals.values())
                if s not in [AUGSYMBOL, STOP])

    def get_production_id(self, name):
        "Returns first production id for the given symbol name"
        for p in self.productions:
            if p.symbol.fqn == name:
                return p.prod_id

    @staticmethod
    def from_struct(productions, start_symbol=None):
        """Used internally to bootstrap grammar file parser."""
        productions, terminals = create_productions_terminals(productions)
        return Grammar(productions,
                       terminals=terminals,
                       start_symbol=start_symbol)

    @staticmethod
    def _parse(parse_fun_name, what_to_parse, recognizers=None,
               ignore_case=False, re_flags=re.MULTILINE, debug=False,
               debug_parse=False, debug_colors=False,
               _no_check_recognizers=False):
        from .parser import Context
        context = Context()
        context.extra = extra = GrammarContext()
        extra.re_flags = re_flags
        extra.ignore_case = ignore_case
        extra.debug = debug
        extra.debug_colors = debug_colors
        extra.classes = {}
        extra.inline_terminals = {}
        extra.imported_with = None
        extra.grammar = None
        grammar_parser = get_grammar_parser(debug_parse, debug_colors)
        imports, productions, terminals, classes = \
            getattr(grammar_parser, parse_fun_name)(what_to_parse,
                                                    context=context)
        g = Grammar(productions=productions,
                    terminals=terminals,
                    classes=classes,
                    imports=imports,
                    recognizers=recognizers,
                    file_path=what_to_parse
                    if parse_fun_name == 'parse_file' else None,
                    _no_check_recognizers=_no_check_recognizers)
        termui.colors = debug_colors
        if debug:
            g.print_debug()

        return g

    @staticmethod
    def from_string(grammar_str, **kwargs):
        return Grammar._parse('parse', grammar_str, **kwargs)

    @staticmethod
    def from_file(file_name, **kwargs):
        file_name = path.realpath(file_name)
        return Grammar._parse('parse_file', file_name, **kwargs)

    def print_debug(self):
        a_print("*** GRAMMAR ***", new_line=True)
        h_print("Terminals:")
        prints(" ".join([text(t) for t in self.terminals]))
        h_print("NonTerminals:")
        prints(" ".join([text(n) for n in self.nonterminals]))

        h_print("Productions:")
        for p in self.productions:
            prints(text(p))


class PGFileImport(object):
    """
    Represents import of a grammar file.

    Attributes:
    module_name (str): Name of this import. By default is the name of grammar
        file without .pg extension.
    file_path (str): A canonical full path of the imported .pg file.
    context (Context): The parsing context.
    imported_with (PGFileImport): First import this import is imported from.
        Used for FQN calculation.
    grammar (Grammar): Grammar object under construction.
    pgfile (PGFile instance or None):

    """
    def __init__(self, module_name, file_path, context):
        self.module_name = module_name
        self.file_path = file_path
        self.context = context
        self.imported_with = context.extra.imported_with
        self.grammar = None
        self.pgfile = None

    @property
    def fqn(self):
        "A fully qualified name of the import following the first import path."
        if self.imported_with:
            return "{}.{}".format(self.imported_with.fqn, self.module_name)
        else:
            return self.module_name

    def load_pgfile(self):
        if self.pgfile is None:
            # First search the global registry of imported files.
            if self.file_path in self.grammar.imported_files:
                self.pgfile = self.grammar.imported_files[self.file_path]
            else:
                # If not found construct new PGFile
                from .parser import Context
                context = Context(extra=self.context.extra,
                                  file_name=self.file_path)
                context.extra.inline_terminals = {}
                context.extra.imported_with = self
                imports, productions, terminals, classes = \
                    get_grammar_parser(
                        self.context.extra.debug,
                        self.context.extra.debug_colors).parse_file(
                            self.file_path, context=context)
                self.pgfile = PGFile(productions=productions,
                                     terminals=terminals,
                                     classes=classes,
                                     imports=imports,
                                     grammar=self.grammar,
                                     imported_with=self,
                                     file_path=self.file_path)

    def resolve_symbol_by_name(self, symbol_name, location=None):
        "Resolves symbol from the imported file."

        return self.pgfile.resolve_symbol_by_name(symbol_name, location)

    def resolve_action_by_name(self, action_name):
        "Resolves action from the imported file."

        return self.pgfile.resolve_action_by_name(action_name)


def create_productions_terminals(productions):
    """Creates Production instances from the list of productions given in
    the form:
    [LHS, RHS, optional ASSOC, optional PRIOR].
    Where LHS is grammar symbol and RHS is a list or tuple of grammar
    symbols from the right-hand side of the production.
    """
    gp = []
    inline_terminals = {}
    for p in productions:
        assoc = ASSOC_NONE
        prior = DEFAULT_PRIORITY
        symbol = p[0]
        if not isinstance(symbol, NonTerminal):
            raise GrammarError(
                location=None,
                message="Invalid production symbol '{}' "
                "for production '{}'".format(symbol, text(p)))
        rhs = ProductionRHS(p[1])
        if len(p) > 2:
            assoc = p[2]
        if len(p) > 3:
            prior = p[3]

        # Convert strings to string recognizers
        for idx, t in enumerate(rhs):
            if isinstance(t, text):
                if t not in inline_terminals:
                    inline_terminals[t] = \
                        Terminal(recognizer=StringRecognizer(t), name=t)
                rhs[idx] = Reference(location=None, name=t)
            elif isinstance(t, Terminal):
                if t.name not in inline_terminals:
                    inline_terminals[t.name] = t
                rhs[idx] = Reference(location=None, name=t.name)

        gp.append(Production(symbol, rhs, assoc=assoc, prior=prior))

    return gp, list(inline_terminals.values())


def make_multiplicity_name(symbol_name, multiplicity=None,
                           separator_name=None):
    if multiplicity is None or multiplicity == MULT_ONE:
        return symbol_name
    name_by_mult = {
        MULT_ZERO_OR_MORE: "0",
        MULT_ONE_OR_MORE: "1",
        MULT_OPTIONAL: "opt"
    }
    if multiplicity:
        return "{}_{}{}".format(
            symbol_name, name_by_mult[multiplicity],
            "_{}".format(separator_name) if separator_name else "")


def check_name(context, name):
    """
    Used in actions to check for reserved names usage.
    """

    if name in RESERVED_SYMBOL_NAMES:
        raise GrammarError(
            location=Location(context),
            message='Rule name "{}" is reserved.'.format(name))
    if '.' in name:
        raise GrammarError(
            location=Location(context),
            message='Using dot in names is not allowed.'.format(name))


class GrammarContext:
    def __deepcopy__(self, memo):
        # Use shallow copy for grammar context
        return copy(self)


# Grammar for grammars

(PGFILE,
 IMPORTS,
 IMPORT,
 PRODUCTION_RULES,
 PRODUCTION_RULE,
 PRODUCTION_RULE_WITH_ACTION,
 PRODUCTION_RULE_RHS,
 PRODUCTION,
 TERMINAL_RULES,
 TERMINAL_RULE,
 TERMINAL_RULE_WITH_ACTION,
 PROD_META_DATA,
 PROD_META_DATAS,
 TERM_META_DATA,
 TERM_META_DATAS,
 USER_META_DATA,
 CONST,

 ASSIGNMENT,
 ASSIGNMENTS,
 PLAIN_ASSIGNMENT,
 BOOL_ASSIGNMENT,

 GSYMBOL_REFERENCE,
 OPT_REP_OPERATOR,
 REP_OPERATOR_ZERO,
 REP_OPERATOR_ONE,
 REP_OPERATOR_OPTIONAL,
 OPT_REP_MODIFIERS_EXP,
 OPT_REP_MODIFIERS,
 OPT_REP_MODIFIER,

 GSYMBOL,
 RECOGNIZER,
 LAYOUT,
 LAYOUT_ITEM,
 COMMENT,
 CORNC,
 CORNCS) = [NonTerminal(name) for name in [
     'PGFile',
     'Imports',
     'Import',
     'ProductionRules',
     'ProductionRule',
     'ProductionRuleWithAction',
     'ProductionRuleRHS',
     'Production',
     'TerminalRules',
     'TerminalRule',
     'TerminalRuleWithAction',
     'ProductionMetaData',
     'ProductionMetaDatas',
     'TerminalMetaData',
     'TerminalMetaDatas',
     'UserMetaData',
     'Const',

     'Assignment',
     'Assignments',
     'PlainAssignment',
     'BoolAssignment',

     'GrammarSymbolReference',
     'OptRepeatOperator',
     'RepeatOperatorZero',
     'RepeatOperatorOne',
     'RepeatOperatorOptional',
     'OptionalRepeatModifiersExpression',
     'OptionalRepeatModifiers',
     'OptionalRepeatModifier',

     'GrammarSymbol',
     'Recognizer',
     'LAYOUT',
     'LAYOUT_ITEM',
     'Comment',
     'CORNC',
     'CORNCS']]

pg_terminals = \
    (NAME,
     REGEX_TERM,
     INT_CONST,
     FLOAT_CONST,
     BOOL_CONST,
     STR_CONST,
     ACTION,
     WS,
     COMMENTLINE,
     NOTCOMMENT) = [Terminal(name, RegExRecognizer(regex)) for name, regex in
                    [
                        ('Name', r'[a-zA-Z_][a-zA-Z0-9_\.]*'),
                        ('RegExTerm', r'\/(\\.|[^\/\\])*\/'),
                        ('IntConst', r'\d+'),
                        ('FloatConst',
                         r'''[+-]?(\d+\.\d*|\.\d+)([eE][+-]?\d+)?(?<=[\w\.])(?![\w\.])'''),  # noqa
                        ('BoolConst', r'true|false'),
                        ('StrConst', r'''(?s)('[^'\\]*(?:\\.[^'\\]*)*')|'''
                         r'''("[^"\\]*(?:\\.[^"\\]*)*")'''),
                        ('Action', r'@[a-zA-Z0-9_]+'),
                        ('WS', r'\s+'),
                        ('CommentLine', r'\/\/.*'),
                        ('NotComment', r'((\*[^\/])|[^\s*\/]|\/[^\*])+'),
                    ]]

pg_productions = [
    [PGFILE, [PRODUCTION_RULES, EOF]],
    [PGFILE, [IMPORTS, PRODUCTION_RULES, EOF]],
    [PGFILE, [PRODUCTION_RULES, 'terminals', TERMINAL_RULES, EOF]],
    [PGFILE, [IMPORTS, PRODUCTION_RULES, 'terminals', TERMINAL_RULES, EOF]],
    [PGFILE, ['terminals', TERMINAL_RULES, EOF]],
    [IMPORTS, [IMPORTS, IMPORT]],
    [IMPORTS, [IMPORT]],
    [IMPORT, ['import', STR_CONST, ';']],
    [IMPORT, ['import', STR_CONST, 'as', NAME, ';']],
    [PRODUCTION_RULES, [PRODUCTION_RULES, PRODUCTION_RULE_WITH_ACTION]],
    [PRODUCTION_RULES, [PRODUCTION_RULE_WITH_ACTION]],

    [PRODUCTION_RULE_WITH_ACTION, [ACTION, PRODUCTION_RULE]],
    [PRODUCTION_RULE_WITH_ACTION, [PRODUCTION_RULE]],
    [PRODUCTION_RULE, [NAME, ':', PRODUCTION_RULE_RHS, ';']],
    [PRODUCTION_RULE, [NAME, '{', PROD_META_DATAS, '}', ':',
                       PRODUCTION_RULE_RHS, ';']],
    [PRODUCTION_RULE_RHS, [PRODUCTION_RULE_RHS, '|', PRODUCTION],
     ASSOC_LEFT, 5],
    [PRODUCTION_RULE_RHS, [PRODUCTION], ASSOC_LEFT, 5],
    [PRODUCTION, [ASSIGNMENTS]],
    [PRODUCTION, [ASSIGNMENTS, '{', PROD_META_DATAS, '}']],

    [TERMINAL_RULES, [TERMINAL_RULES, TERMINAL_RULE_WITH_ACTION]],
    [TERMINAL_RULES, [TERMINAL_RULE_WITH_ACTION]],
    [TERMINAL_RULE_WITH_ACTION, [ACTION, TERMINAL_RULE]],
    [TERMINAL_RULE_WITH_ACTION, [TERMINAL_RULE]],
    [TERMINAL_RULE, [NAME, ':', RECOGNIZER, ';'], ASSOC_LEFT, 15],
    [TERMINAL_RULE, [NAME, ':', ';'], ASSOC_LEFT, 15],
    [TERMINAL_RULE, [NAME, ':', RECOGNIZER, '{', TERM_META_DATAS, '}', ';'],
     ASSOC_LEFT, 15],
    [TERMINAL_RULE, [NAME, ':', '{', TERM_META_DATAS, '}', ';'],
     ASSOC_LEFT, 15],

    [PROD_META_DATA, ['left']],
    [PROD_META_DATA, ['reduce']],
    [PROD_META_DATA, ['right']],
    [PROD_META_DATA, ['shift']],
    [PROD_META_DATA, ['dynamic']],
    [PROD_META_DATA, ['nops']],   # no prefer shifts
    [PROD_META_DATA, ['nopse']],  # no prefer shifts over empty
    [PROD_META_DATA, [INT_CONST]],  # priority
    [PROD_META_DATA, [USER_META_DATA]],
    [PROD_META_DATAS, [PROD_META_DATAS, ',', PROD_META_DATA], ASSOC_LEFT],
    [PROD_META_DATAS, [PROD_META_DATA]],

    [TERM_META_DATA, ['prefer']],
    [TERM_META_DATA, ['finish']],
    [TERM_META_DATA, ['nofinish']],
    [TERM_META_DATA, ['dynamic']],
    [TERM_META_DATA, [INT_CONST]],  # priority
    [TERM_META_DATA, [USER_META_DATA]],
    [TERM_META_DATAS, [TERM_META_DATAS, ',', TERM_META_DATA]],
    [TERM_META_DATAS, [TERM_META_DATA]],

    # User custom meta-data
    [USER_META_DATA, [NAME, ':', CONST]],
    [CONST, [INT_CONST]],
    [CONST, [FLOAT_CONST]],
    [CONST, [BOOL_CONST]],
    [CONST, [STR_CONST]],

    # Assignments
    [ASSIGNMENT, [PLAIN_ASSIGNMENT]],
    [ASSIGNMENT, [BOOL_ASSIGNMENT]],
    [ASSIGNMENT, [GSYMBOL_REFERENCE]],
    [ASSIGNMENTS, [ASSIGNMENTS, ASSIGNMENT]],
    [ASSIGNMENTS, [ASSIGNMENT]],
    [PLAIN_ASSIGNMENT, [NAME, '=', GSYMBOL_REFERENCE]],
    [BOOL_ASSIGNMENT, [NAME, '?=', GSYMBOL_REFERENCE]],

    # Regex-like repeat operators
    [GSYMBOL_REFERENCE, [GSYMBOL, OPT_REP_OPERATOR]],
    [OPT_REP_OPERATOR, [REP_OPERATOR_ZERO]],
    [OPT_REP_OPERATOR, [REP_OPERATOR_ONE]],
    [OPT_REP_OPERATOR, [REP_OPERATOR_OPTIONAL]],
    [OPT_REP_OPERATOR, [EMPTY]],
    [REP_OPERATOR_ZERO, ['*', OPT_REP_MODIFIERS_EXP]],
    [REP_OPERATOR_ONE, ['+', OPT_REP_MODIFIERS_EXP]],
    [REP_OPERATOR_OPTIONAL, ['?', OPT_REP_MODIFIERS_EXP]],
    [OPT_REP_MODIFIERS_EXP, ['[', OPT_REP_MODIFIERS, ']']],
    [OPT_REP_MODIFIERS_EXP, [EMPTY]],
    [OPT_REP_MODIFIERS, [OPT_REP_MODIFIERS, ',', OPT_REP_MODIFIER]],
    [OPT_REP_MODIFIERS, [OPT_REP_MODIFIER]],
    [OPT_REP_MODIFIER, [NAME]],

    [GSYMBOL, [NAME]],
    [GSYMBOL, [STR_CONST]],
    [RECOGNIZER, [STR_CONST]],
    [RECOGNIZER, [REGEX_TERM]],

    # Support for comments,
    [LAYOUT, [LAYOUT_ITEM]],
    [LAYOUT, [LAYOUT, LAYOUT_ITEM]],
    [LAYOUT_ITEM, [WS]],
    [LAYOUT_ITEM, [COMMENT]],
    [LAYOUT_ITEM, [EMPTY]],
    [COMMENT, ['/*', CORNCS, '*/']],
    [COMMENT, [COMMENTLINE]],
    [CORNCS, [CORNC]],
    [CORNCS, [CORNCS, CORNC]],
    [CORNCS, [EMPTY]],
    [CORNC, [COMMENT]],
    [CORNC, [NOTCOMMENT]],
    [CORNC, [WS]]
]


grammar_parser = None


def get_grammar_parser(debug, debug_colors):
    global grammar_parser
    if not grammar_parser:
        from parglare import Parser
        grammar_parser = Parser(Grammar.from_struct(pg_productions, PGFILE),
                                actions=pg_actions,
                                debug=debug,
                                debug_colors=debug_colors)
    EMPTY.action = pass_none
    EOF.action = pass_none
    return grammar_parser


def act_pgfile(context, nodes):
    imports, productions, terminals = [], [], []
    while nodes:
        first = nodes.pop(0)
        if first and type(first) is list:
            if type(first[0]) is PGFileImport:
                imports = first
            elif type(first[0]) is Production:
                productions = first
            elif type(first[0]) is Terminal:
                terminals = first

    for terminal in context.extra.inline_terminals.values():
        terminals.append(terminal)

    return [imports, productions, terminals, context.extra.classes]


def act_import(context, nodes):
    if not context.file_name:
        raise GrammarError(location=Location(context),
                           message='Import can be used only for grammars '
                           'defined in files.')
    import_path = nodes[1]
    module_name = nodes[3] if len(nodes) > 3 else None
    if module_name is None:
        module_name = path.splitext(path.basename(import_path))[0]
    if not path.isabs(import_path):
        import_path = path.realpath(path.join(path.dirname(context.file_name),
                                              import_path))
    else:
        import_path = path.realpath(import_path)

    return PGFileImport(module_name, import_path, context)


def act_production_rules(_, nodes):
    e1, e2 = nodes
    e1.extend(e2)
    return e1


def act_production_rule_with_action(_, nodes):
    if len(nodes) > 1:
        action_name, productions = nodes
        # Strip @ char
        action_name = action_name[1:]
        for p in productions:
            p.symbol.action_name = action_name
    else:
        productions = nodes[0]

    return productions


def act_production_rule(context, nodes):
    if len(nodes) == 4:
        # No meta-data
        name, _, rhs_prods, __ = nodes
        rule_meta_datas = {}
    else:
        name, rule_meta_datas, rhs_prods = nodes[0], nodes[2], nodes[5]
        rule_meta_datas = get_production_rule_meta_datas(rule_meta_datas)

    check_name(context, name)

    symbol = NonTerminal(name, location=Location(context),
                         imported_with=context.extra.imported_with,
                         user_meta=rule_meta_datas.get('user_meta', None))

    # Collect all productions for this rule
    prods = []
    attrs = {}
    for prod in rhs_prods:
        assignments, meta_datas = prod
        # Here we know the indexes of assignments
        for idx, a in enumerate(assignments):
            if a.name:
                a.index = idx
        gsymbols = (a.symbol for a in assignments)
        assoc = meta_datas.get('assoc', rule_meta_datas.get('assoc',
                                                            ASSOC_NONE))
        prior = meta_datas.get('priority',
                               rule_meta_datas.get('priority',
                                                   DEFAULT_PRIORITY))
        dynamic = meta_datas.get('dynamic',
                                 rule_meta_datas.get('dynamic', False))
        nops = meta_datas.get('nops',
                              rule_meta_datas.get('nops', False))
        nopse = meta_datas.get('nopse', rule_meta_datas.get('nopse', False))

        # User meta-data if formed by rule-level user meta-data with overrides
        # from production-level user meta-data.
        user_meta = dict(rule_meta_datas.get('user_meta', {}))
        user_meta.update(meta_datas.get('user_meta', {}))
        prods.append(Production(symbol,
                                ProductionRHS(gsymbols),
                                assignments=assignments,
                                assoc=assoc,
                                prior=prior,
                                dynamic=dynamic,
                                nops=nops,
                                nopse=nopse,
                                user_meta=user_meta))

        for a in assignments:
            if a.name:
                attrs[a.name] = PGAttribute(a.name, a.multiplicity,
                                            a.symbol_name)
            # TODO: check/handle multiple assignments to the same attribute
            #       If a single production have multiple assignment of the
            #       same attribute, multiplicity must be set to many.

    # If named matches are used create Python class that will be used
    # for object instantiation.
    if attrs:
        class ParglareMetaClass(type):

            def __repr__(cls):
                return '<parglare:{} class at {}>'.format(name, id(cls))

        @add_metaclass(ParglareMetaClass)
        class ParglareClass(object):
            """Dynamically created class. Each parglare rule that uses named
            matches by default uses this action that will create Python object
            of this class.

            Attributes:
                _pg_attrs(dict): A dict of meta-attributes keyed by name.
                    Used by common rules.
                _pg_position(int): A position in the input string where
                    this class is defined.
                _pg_position_end(int): A position in the input string where
                    this class ends.

            """

            _pg_attrs = attrs

            def __init__(self, **attrs):
                for attr_name, attr_value in attrs.items():
                    setattr(self, attr_name, attr_value)

            def __repr__(self):
                if hasattr(self, 'name'):
                    return "<{}:{}>".format(name, self.name)
                else:
                    return "<parglare:{} instance at {}>"\
                        .format(name, hex(id(self)))

        ParglareClass.__name__ = str(symbol.fqn)
        if symbol.fqn in context.extra.classes:
            # If rule has multiple definition merge attributes.
            context.extra.classes[symbol.fqn]._pg_attrs.update(attrs)
        else:
            context.extra.classes[symbol.fqn] = ParglareClass

        symbol.action_name = 'obj'

    return prods


def get_production_rule_meta_datas(raw_meta_datas):
    meta_datas = {}
    for meta_data in raw_meta_datas:
        if meta_data in ['left', 'reduce']:
            meta_datas['assoc'] = ASSOC_LEFT
        elif meta_data in ['right', 'shift']:
            meta_datas['assoc'] = ASSOC_RIGHT
        elif meta_data == 'dynamic':
            meta_datas['dynamic'] = True
        elif meta_data == 'nops':
            meta_datas['nops'] = True
        elif meta_data == 'nopse':
            meta_datas['nopse'] = True
        elif type(meta_data) is int:
            meta_datas['priority'] = meta_data
        else:
            # User meta-data
            assert type(meta_data) is list
            name, _, value = meta_data
            meta_datas.setdefault('user_meta', {})[name] = value
    return meta_datas


def act_production(_, nodes):
    assignments = nodes[0]
    meta_datas = {}
    if len(nodes) > 1:
        meta_datas = get_production_rule_meta_datas(nodes[2])

    return (assignments, meta_datas)


def _set_term_props(term, props):
    for t in props:
        if type(t) is int:
            term.prior = t
        elif type(t) is list:
            # User meta-data
            name, _, value = t
            term.add_user_meta_data(name, value)
        elif t == 'finish':
            term.finish = True
        elif t == 'nofinish':
            term.finish = False
        elif t == 'prefer':
            term.prefer = True
        elif t == 'dynamic':
            term.dynamic = True
        else:
            print(t)
            assert False


def act_term_rule(context, nodes):

    name = nodes[0]
    recognizer = nodes[2]

    check_name(context, name)
    term = Terminal(name, recognizer, location=Location(context),
                    imported_with=context.extra.imported_with)
    if len(nodes) > 4:
        _set_term_props(term, nodes[4])
    return term


def act_term_rule_empty_body(context, nodes):
    name = nodes[0]

    check_name(context, name)
    term = Terminal(name, location=Location(context),
                    imported_with=context.extra.imported_with)
    term.recognizer = None
    if len(nodes) > 3:
        _set_term_props(term, nodes[3])
    return term


def act_term_rule_with_action(context, nodes):
    if len(nodes) > 1:
        action_name, term = nodes
        # Strip @ char
        action_name = action_name[1:]
        term.action_name = action_name
    else:
        term = nodes[0]

    return term


def act_gsymbol_reference(context, nodes):
    """Repetition operators (`*`, `+`, `?`) will create additional productions in
    the grammar with name generated from original symbol name and suffixes:
    - `_0` - for `*`
    - `_1` - for `+`
    - `_opt` - for `?`

    Zero or more produces `one or more` productions and additional productions
    of the form:

    ```
    somerule_0: somerule_1 | EMPTY;
    ```

    In addition if separator is used another suffix is added which is the name
    of the separator rule, for example:

    ```
    spam*[comma] --> spam_0_comma and spam_1_comma
    spam+[comma] --> spam_1_comma
    spam* --> spam_0 and spam_1
    spam? --> spam_opt
    ```

    """
    symbol_ref, rep_op = nodes

    if rep_op:

        if len(rep_op) > 1:
            rep_op, modifiers = rep_op
        else:
            rep_op = rep_op[0]
            modifiers = None

        sep_ref = None
        if modifiers:
            sep_ref = modifiers[1]
            sep_ref = Reference(Location(context), sep_ref)
            symbol_ref.separator = sep_ref

        if rep_op == '*':
            symbol_ref.multiplicity = MULT_ZERO_OR_MORE
        elif rep_op == '+':
            symbol_ref.multiplicity = MULT_ONE_OR_MORE
        else:
            symbol_ref.multiplicity = MULT_OPTIONAL

    return symbol_ref


def act_gsymbol_string_recognizer(context, nodes):
    recognizer = act_recognizer_str(context, nodes)

    terminal_ref = Reference(Location(context), recognizer.name)

    if terminal_ref.name not in context.extra.inline_terminals:
        check_name(context, terminal_ref.name)
        context.extra.inline_terminals[terminal_ref.name] = \
            Terminal(terminal_ref.name, recognizer, location=Location(context))

    return terminal_ref


def act_assignment(_, nodes):
    gsymbol_reference = nodes[0]
    if type(gsymbol_reference) is list:
        # Named match
        name, op, gsymbol_reference = gsymbol_reference
    else:
        name, op = None, None

    return Assignment(name, op, gsymbol_reference)


def act_recognizer_str(context, nodes):
    value = nodes[0]
    value = value.replace(r'\"', '"')\
                 .replace(r"\'", "'")\
                 .replace(r"\\", "\\")\
                 .replace(r"\n", "\n")\
                 .replace(r"\t", "\t")
    return StringRecognizer(value, ignore_case=context.extra.ignore_case)


def act_recognizer_regex(context, nodes):
    value = nodes[0]
    return RegExRecognizer(value, re_flags=context.extra.re_flags,
                           ignore_case=context.extra.ignore_case)


def act_str_term(context, value):
    value = value[1:-1]
    value = value.replace(r"\\", "\\")
    value = value.replace(r"\'", "'")
    return value


def act_regex_term(context, value):
    return value[1:-1]


pg_actions = {
    "PGFile": act_pgfile,
    "Imports": collect,
    "Import": act_import,

    "ProductionRules": [act_production_rules, pass_single],
    'ProductionRule': act_production_rule,
    'ProductionRuleWithAction': act_production_rule_with_action,
    'ProductionRuleRHS': collect_sep,
    'Production': act_production,

    'TerminalRules': collect,
    'TerminalRule': [act_term_rule,
                     act_term_rule_empty_body,
                     act_term_rule,
                     act_term_rule_empty_body],
    'TerminalRuleWithAction': act_term_rule_with_action,

    "ProductionMetaDatas": collect_sep,
    "TerminalMetaDatas": collect_sep,

    "Assignment": act_assignment,
    "Assignments": collect,

    'GrammarSymbolReference': act_gsymbol_reference,

    'GrammarSymbol': [lambda context, nodes: Reference(Location(context),
                                                       nodes[0]),
                      act_gsymbol_string_recognizer],

    'Recognizer': [act_recognizer_str, act_recognizer_regex],

    'StrConst': act_str_term,
    'RegExTerm': act_regex_term,

    # Constants
    'IntConst': lambda _, value: int(value),
    'FloatConst': lambda _, value: float(value),
    'BoolConst': lambda _, value: value and value.lower() == 'true',

}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest
from os.path import join, dirname
from parglare import Parser, Grammar
from .expression_grammar import get_grammar
from parglare.exceptions import ParseError


def test_parsing():
    grammar = get_grammar()
    p = Parser(grammar)
    assert p.parse("id+id+id")


def test_parsing_from_file():
    grammar = get_grammar()
    p = Parser(grammar)
    assert p.parse_file(join(dirname(__file__), 'parsing_from_file.txt'))


def test_partial_parse():
    """
    Not giving EOF at the end of the sequence enables parsing of the beginning
    of the input string.
    """
    grammar = """
    S: 'a' B;
    B: 'b';
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)

    # Parser should succesfuly parse 'ab' at the beggining.
    parser.parse('abc')

    # But if EOF is given it will match only at the end of the string,
    # thus, the whole string must be parsed in order for parsing to
    # succeed.
    grammar = """
    S: 'a' B EOF;
    B: 'b';
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    parser.parse('a b')
    with pytest.raises(ParseError):
        parser.parse('a b c')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
from parglare.parser import Parser
from .expression_grammar import get_grammar


def test_tree_str():

    grammar = get_grammar()
    p = Parser(grammar, build_tree=True)

    res = p.parse("""id+  id * (id
    +id  )
    """)

    ts = res.tree_str()

    assert '+[18->19, "+"]' in ts
    assert ')[23->24, ")"]' in ts
    assert 'F[10->24]' in ts



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import sys
import pytest  # noqa
from parglare import Grammar, Parser, GLRParser, EMPTY, EOF
from parglare.tables import first, follow, create_table, SHIFT, REDUCE
from .expression_grammar import OPEN, ID, T, E, MULT, CLOSE, PLUS, get_grammar

HAS_MOCK = sys.version_info[0] >= 3
if HAS_MOCK:
    from unittest.mock import patch


def test_first():
    """
    Tests FIRST function.

    FIRST function returns a set of terminals that could start the sentence
    derived from the given non-terminal.
    """

    expression_grammar = get_grammar()
    first_set = first(expression_grammar)

    assert OPEN in first_set[T]
    assert ID in first_set[T]


def test_first_empty_in_rhs():
    """
    Test FIRST calculation when there are empty derivations in RHS of a
    production.
    """

    grammar = """
    S: A C;
    A: B | EMPTY;

    terminals
    B: "b";
    C: "c";
    """

    g = Grammar.from_string(grammar)

    first_set = first(g)

    A = g.get_nonterminal('A')
    B = g.get_terminal('B')
    C = g.get_terminal('C')
    S = g.get_nonterminal('S')

    assert EMPTY in first_set[A]
    assert B in first_set[A]

    assert B in first_set[S]

    # 'A' can derive empty, thus 'C' must be in firsts of 'S'.
    assert C in first_set[S]


def test_follow():
    """Tests FOLLOW function.

    FOLLOW function calculates a set of terminals that can follow each grammar
    non-terminal in any of derived sentential forms.
    """

    expression_grammar = get_grammar()
    follow_set = follow(expression_grammar)

    assert follow_set[E] == set([CLOSE, PLUS, EOF])

    # Follow of T must contain all of follow of E
    assert follow_set[T] == set([MULT, CLOSE, PLUS, EOF])


def test_table_construction():
    """
    Tests LR table construction.
    """
    # From the Knuth's 1965 paper: On the Translation of Languages from Left to
    # Right
    grammar = r"""
    S: 'a' A 'd' | 'b' A 'd';
    A: 'c' A | 'c';
    """

    g = Grammar.from_string(grammar)
    table = create_table(g)

    c = g.get_terminal('c')
    d = g.get_terminal('d')

    assert len(table.states) == 11
    assert table.states[0].symbol.name == "S'"
    state = table.states[2]
    assert state.symbol.name == 'a'
    assert len(state.kernel_items) == 1
    assert len(state.items) == 3
    assert len(state.actions) == 1
    assert len(state.actions[c]) == 1
    action = list(state.actions.values())[0][0]
    assert action.action == SHIFT
    assert action.state.state_id == 6

    state = table.states[6]
    assert state.symbol.name == 'c'
    assert len(state.kernel_items) == 2
    assert len(state.items) == 4
    assert len(state.actions) == 2
    assert len(state.actions[c]) == 1
    assert len(state.actions[d]) == 1
    action = list(state.actions.values())[0][0]
    assert action.action == REDUCE
    assert action.prod.prod_id == 4
    action = list(state.actions.values())[1][0]
    assert action.action == SHIFT
    assert action.state.state_id == 6


def test_associativity_conflicts_resolving():
    """
    Test that using associativity will resolve conflicts.
    """
    grammar = r"""
    E: E '+' E | number;

    terminals
    number: /d+/;
    """

    g = Grammar.from_string(grammar)
    table = create_table(g)

    assert len(table.sr_conflicts) == 1

    grammar = r"""
    E: E '+' E {left} | number;

    terminals
    number: /d+/;
    """

    g = Grammar.from_string(grammar)
    table = create_table(g)

    assert len(table.sr_conflicts) == 0


def test_priority_conflicts_resolving():
    """
    Test that using priority will resolve conflicts.
    """


def test_prefer_shifts_no_sr_conflicts():
    """
    Test that grammar with S/R conflict will be resolved to SHIFT actions
    if prefer_shift option is used.
    """
    # This grammar has S/R conflict as B+ may consume multiple single "a" A
    # because "b" is optional. Thus, parser can't decide if it should shift "a"
    # or reduce by 'B: "b"? A+' and later by 'S: B+'; Most of the time we want
    # gready behavior so in case of doubt parser will choose shift if
    # prefer_shift is set to `True`. This means that the parser will first
    # consume all "a" using A+ and that reduce B at the end.
    grammar = r"""
    S: B+ EOF;
    B: "b"? A+;

    terminals
    A: "a";
    """
    g = Grammar.from_string(grammar)

    # There is a shift reduce conflict so we can't use LR parser.
    table = create_table(g)
    assert len(table.sr_conflicts) == 1

    # But we can eliminate conflict by prefer_shifts strategy.
    table = create_table(g, prefer_shifts=True)
    assert len(table.sr_conflicts) == 0

    # With prefer_shifts we get a greedy behavior
    input_str = 'b a a a b a a'
    output = [[['b', ['a', 'a', 'a']], ['b', ['a', 'a']]], None]
    parser = Parser(g, prefer_shifts=True)
    result = parser.parse(input_str)
    assert result == output

    # GLR parser can parse without prefer_shifts strategy. This grammar is
    # ambiguous but, in this case, implicit elimination of empty trees (due to
    # optional "b") will yield a single result.
    parser = GLRParser(g)
    result = parser.parse(input_str)
    assert len(result) == 1
    assert result[0] == output

    parser = GLRParser(g, prefer_shifts=True)
    result = parser.parse(input_str)
    assert len(result) == 1
    assert result[0] == output


def test_prefer_shifts_over_empty_reductions():
    """
    Test strategy that will choose SHIFT when in conflict with EMPTY reduction.
    """
    # TODO


def test_precomputed_table():
    """If parser is initialized with a `table` parameter then this table
    should be used, and no call to create_table should be made."""
    grammar = get_grammar()
    table = create_table(grammar)
    if HAS_MOCK:
        with patch('parglare.tables.create_table') as mocked_create_table:
            parser = GLRParser(grammar, table=table)
            assert not mocked_create_table.called
    else:
        parser = GLRParser(grammar, table=table)
    parser.parse('id+id')



import os
from parglare import Grammar, Parser

this_folder = os.path.dirname(__file__)

model_str = '''
modelID 42
component myComponent {
    in SomeInputSlot
    out SomeOutputSlot
}
'''


def test_imported_actions_connect_by_symbol_name():
    g = Grammar.from_file(os.path.join(this_folder, 'by_symbol_name/model.pg'))
    model = Parser(g).parse(model_str)
    # Check that base.pg actions are properly loaded and triggered.
    assert model.modelID == 42


def test_imported_actions_connect_by_action_name():
    g = Grammar.from_file(os.path.join(this_folder, 'by_action_name/model.pg'))
    model = Parser(g).parse(model_str)
    # Check that base.pg actions are properly loaded and triggered.
    assert model.modelID == 42


def test_imported_actions_connect_by_decorator_action_name():
    g = Grammar.from_file(os.path.join(this_folder,
                                       'by_decorator_action_name/model.pg'))
    model = Parser(g).parse(model_str)
    # Check that base.pg actions are properly loaded and triggered.
    assert model.modelID == 42


def test_imported_actions_override():
    """
    Test that actions loaded from `*_actions.py` files can be overriden by
    users actions.
    """

    # We can override either by fqn of symbol
    g = Grammar.from_file(os.path.join(this_folder, 'by_symbol_name/model.pg'))
    actions = {
        'base.NUMERIC_ID': lambda _, value: 43
    }
    model = Parser(g, actions=actions).parse(model_str)
    assert model.modelID == 43

    # Or by action name used in grammar for the given symbol
    g = Grammar.from_file(os.path.join(this_folder, 'by_action_name/model.pg'))
    actions = {
        'base.number': lambda _, value: 43
    }
    model = Parser(g, actions=actions).parse(model_str)
    assert model.modelID == 43

    # Override by FQN takes precendence
    g = Grammar.from_file(os.path.join(this_folder, 'by_action_name/model.pg'))
    actions = {
        'base.NUMERIC_ID': lambda _, value: 43
    }
    model = Parser(g, actions=actions).parse(model_str)
    assert model.modelID == 43


def test_imported_actions_override_by_grammar_actions():
    """
    Test that actions loaded from `*_actions.py` files can override actions
    imported from other grammar files.
    """

    g = Grammar.from_file(os.path.join(this_folder,
                                       'in_grammar_by_symbol_name/model.pg'))
    model = Parser(g).parse(model_str)
    assert model.modelID == 43

    g = Grammar.from_file(os.path.join(this_folder,
                                       'in_grammar_by_action_name/model.pg'))
    model = Parser(g).parse(model_str)
    assert model.modelID == 43



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
from os.path import join, dirname
from parglare import Grammar, GrammarError


def test_recognizer_for_unexisting_terminal_raises_exception():
    """
    If a recognizer is given for a terminal that can't be found in the grammar
    raise an exception.
    """

    with pytest.raises(GrammarError,
                       match=r'not given for unknown terminal "B".'):
        Grammar.from_file(join(dirname(__file__), 'grammar.pg'))



import pytest   # noqa
import os
import sys
import glob
import imp


def test_examples():

    examples_pat = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                '../../examples/*/*.py')

    # Filter out __init__.py
    examples = [f for f in glob.glob(examples_pat)
                if not any([x in f for x in ['__init__.py',
                                             'molecular',
                                             'custom_table_caching']])]
    for e in examples:
        example_dir = os.path.dirname(e)
        sys.path.insert(0, example_dir)
        (module_name, _) = os.path.splitext(os.path.basename(e))
        (module_file, module_path, desc) = \
            imp.find_module(module_name, [example_dir])

        m = imp.load_module(module_name, module_file, module_path, desc)

        if hasattr(m, 'main'):
            m.main(debug=False)



import os
import time
from parglare import Grammar, Parser

this_folder = os.path.dirname(__file__)


def test_save_load_table():
    """
    Test basic table save/load cycle with table file creation.
    """
    calc_file = os.path.join(this_folder, 'calc.pg')
    variable_file = os.path.join(this_folder, 'variable.pg')
    input_str = 'a = 5   1 + 2 * a - 7'
    input_str_result = 1 + 2 * 5 - 7
    grammar = Grammar.from_file(calc_file)

    table_file = os.path.join(this_folder, 'calc.pgt')
    # remove table file if exists
    try:
        os.remove(table_file)
    except OSError:
        pass

    parser = Parser(grammar)
    assert parser.parse(input_str) == input_str_result

    # Table file must be produced by parser construction.
    assert os.path.exists(table_file)

    last_mtime = os.path.getmtime(table_file)
    time.sleep(1)

    parser = Parser(grammar)

    # Last generated table should be used during parser construction.
    # Currently, it is hard to check this so we'll only check if
    # table_file is not regenerated.
    assert last_mtime == os.path.getmtime(table_file)
    # Parser constructed from persisted table should produce the same result.
    assert parser.parse(input_str) == input_str_result

    # We are now touching variable.pg file
    # This should trigger table file regeneration
    with open(variable_file, 'a'):
        os.utime(variable_file, None)
    parser = Parser(grammar)
    assert parser.parse(input_str) == input_str_result
    # We verify that the table file is newer.
    assert last_mtime < os.path.getmtime(table_file)

    # Now we test that force_load_table will load table even if not
    # newer than the grammar.
    time.sleep(1)
    with open(variable_file, 'a'):
        os.utime(variable_file, None)
    last_mtime = os.path.getmtime(table_file)
    parser = Parser(grammar, force_load_table=True)
    assert last_mtime == os.path.getmtime(table_file)
    parser = Parser(grammar)
    assert last_mtime < os.path.getmtime(table_file)



import os
from parglare import Grammar, Parser

this_folder = os.path.dirname(__file__)


def test_recursive_grammar_import():
    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'))
    assert g

    input_str = '''

    package First
    package Second

    module SomeModule {

        component myComponent {
            in SomeInputSlot
            out SomeOutputSlot
        }


    }

    '''

    result = Parser(g).parse(input_str)
    assert result



import pytest
from parglare import Grammar, GrammarError


def test_terminal_exists_noerror_on_terminal_definition_before():
    """
    Test the situation where we have inline terminal used but the
    same terminal is defined in the `terminals` section.
    """
    grammar = """
    Program: "begin"
             statements=Statement*
             ProgramEnd EOF;
    Statement: "end" "transaction"  // here we are using inline terminal `end`
                                    // instead of using a name reference
             | "command";

    terminals
    ProgramEnd: "end";
    """

    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'match the same string' in str(e)



import pytest  # noqa
from parglare import Grammar, Parser


def test_call_actions_during_tree_build():
    grammar = """
    Program: "begin" MoveCommand* "end";
    MoveCommand: "move" Direction;
    Direction: "up" | "down" | "left" | "right";
    """

    g = Grammar.from_string(grammar)

    code = """
    begin
        move left
        move left
        move up
        move down
    end
    """

    left_moves = []

    def left_dir_collector(_, nodes):
        """Finds all 'left' moves and adds them into a list."""
        term = nodes[0]
        if term.value == "left":
            left_moves.append(term)

    parser = Parser(g, build_tree=True,
                    actions={"Direction": left_dir_collector})
    parser.parse(code)

    # call_actions_during_tree_build is False by default, so left_dir_collector
    # will not be called.
    assert len(left_moves) == 0

    parser.call_actions_during_tree_build = True
    parser.parse(code)

    assert len(left_moves) == 2



import json
from collections import OrderedDict


def table_to_serializable(table):
    """Convert table object to serializable representation composed of
    lists and dicts."""
    # states
    states = []
    for state in table.states:
        states.append(_dump_state(state))

    return states


def save_table(file_name, table):
    with open(file_name, 'w') as f:
        json.dump(table_to_serializable(table), f, sort_keys=True)


def table_from_serializable(serialized_states, grammar):
    """Convert serializable representation of a parsing table into
    LRTable object."""
    from parglare.tables import LRState, LRTable, Action

    states = []
    states_dict = {}
    for json_state in serialized_states:
        state = LRState(grammar, json_state['state_id'],
                        grammar.get_symbol(json_state['symbol']))
        states_dict[state.state_id] = state
        state.finish_flags = json_state['finish_flags']
        state.actions = json_state['actions']
        state.gotos = json_state['gotos']
        states.append(state)

    # Unpack actions and gotos
    for state in states:

        actions = OrderedDict()
        for json_action_fqn in state.actions:
            terminal_fqn, json_actions = json_action_fqn
            term_acts = []
            for json_action in json_actions:
                if 'state_id' in json_action:
                    act_state = states_dict[json_action['state_id']]
                else:
                    act_state = None
                if 'prod_id' in json_action:
                    act_prod = grammar.productions[json_action['prod_id']]
                else:
                    act_prod = None
                term_acts.append(Action(json_action['action'],
                                        act_state, act_prod))

            actions[grammar.get_terminal(terminal_fqn)] = term_acts
        state.actions = actions

        gotos = OrderedDict()
        for json_goto_fqn in state.gotos:
            nonterm_fqn, goto_state = json_goto_fqn
            gotos[grammar.get_nonterminal(nonterm_fqn)] = \
                states_dict[goto_state]
        state.gotos = gotos

    table = LRTable(states, calc_finish_flags=False)

    return table


def load_table(file_name, grammar):
    with open(file_name) as f:
        return table_from_serializable(json.load(f), grammar)


def _dump_state(state):
    s = {}
    s['state_id'] = state.state_id
    s['symbol'] = state.symbol.fqn
    action_items = list(state.actions.items())
    s['actions'] = [[terminal.fqn, _dump_actions(actions)]
                    for terminal, actions in action_items]
    goto_items = list(state.gotos.items())
    s['gotos'] = [[nonterminal.fqn, st.state_id]
                  for nonterminal, st in goto_items]
    s['finish_flags'] = state.finish_flags

    return s


def _dump_actions(actions):
    alist = []
    for action in actions:
        a = {}
        a['action'] = action.action
        if action.state is not None:
            a['state_id'] = action.state.state_id
        if action.prod is not None:
            a['prod_id'] = action.prod.prod_id
        alist.append(a)

    return alist



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest
from parglare import Parser, Grammar
from parglare.grammar import ASSOC_LEFT, ASSOC_RIGHT, DEFAULT_PRIORITY
from parglare.exceptions import GrammarError, ParseError


def test_single_terminal():
    """
    Test that grammar may be just a single terminal.
    """
    grammar = r"""
    S: A;
    terminals
    A: "a";
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    result = parser.parse('a')
    assert result == 'a'

    grammar = r"""
    S: A;
    terminals
    A: /\d+/;
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    result = parser.parse('23')
    assert result == '23'


def test_undefined_grammar_symbol():
    "Tests that undefined grammar symbols raises errors."
    grammar = """
    S: A B;
    A: "a" | B;
    B: id;
    """
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'Unknown symbol' in str(e)
    assert 'id' in str(e)


def test_multiple_terminal_definition():

    grammar = """
    S: A A;
    terminals
    A: "a";
    A: "b";
    """

    with pytest.raises(GrammarError,
                       match=r'.*Multiple definitions of terminal rule.*'):
        Grammar.from_string(grammar)


def test_reserved_symbol_names():
    """
    Test that reserved symbol names can't be used.
    """
    grammar = """
    S: EOF "First";
    EOF: "eof";
    """
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)
    assert 'is reserved' in str(e)

    grammar = """
    S: STOP "First";
    STOP: "stop";
    """
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)
    assert 'is reserved' in str(e)

    grammar = """
    S: EMPTY "First";
    EMPTY: "stop";
    """
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)
    assert 'is reserved' in str(e)


def test_assoc_prior():
    """Test that associativity and priority can be defined for productions and
    terminals.
    """

    grammar = r"""
    E: E '+' E {left, 1};
    E: E '*' E {2, left};
    E: E '^' E {right};
    E: id;
    terminals
    id: /\d+/;
    """

    g = Grammar.from_string(grammar)

    assert g.productions[1].prior == 1
    assert g.productions[1].assoc == ASSOC_LEFT
    assert g.productions[3].assoc == ASSOC_RIGHT
    assert g.productions[3].prior == DEFAULT_PRIORITY

    assert g.productions[3].prior == DEFAULT_PRIORITY

    # Repeat the same but for alternative keywords "shift" and "reduce"
    grammar = r"""
    E: E '+' E {reduce, 1};
    E: E '*' E {2, reduce};
    E: E '^' E {shift};
    E: id;
    terminals
    id: /\d+/;
    """

    g = Grammar.from_string(grammar)

    assert g.productions[1].prior == 1
    assert g.productions[1].assoc == ASSOC_LEFT
    assert g.productions[3].assoc == ASSOC_RIGHT
    assert g.productions[3].prior == DEFAULT_PRIORITY

    assert g.productions[3].prior == DEFAULT_PRIORITY


def test_terminal_priority():
    "Terminals might define priority which is used for lexical disambiguation."

    grammar = """
    S: A | B;
    A: 'a' {15};
    B: 'b';
    """

    g = Grammar.from_string(grammar)

    for t in g.terminals.values():
        if t.name == 'A':
            assert t.prior == 15
        else:
            assert t.prior == DEFAULT_PRIORITY


def test_no_terminal_associavitity():
    "Tests that terminals can't have associativity defined."
    grammar = """
    S: A | B;
    terminals
    A: 'a' {15, left};
    B: 'b';
    """

    with pytest.raises(ParseError) as e:
        Grammar.from_string(grammar)

    assert 'Expected: : but found <NotComment(};)> or <}(})>' \
        in str(e)


def test_terminal_empty_body():
    """
    Test that terminals may have empty bodies (when defined using
    recognizers)
    """
    grammar = """
    S: A | B;
    terminals
    A: {15};
    B: ;
    """

    g = Grammar.from_string(grammar, recognizers={'B': None, 'A': None})

    a = g.get_terminal('A')
    assert a.prior == 15
    b = g.get_terminal('B')
    assert b.recognizer is None


def test_terminal_regexp_with_backslash():
    """Regexp terminals can contain (escaped) backslash."""
    grammar = Grammar.from_string(r"""
    start: t1 t2;
    terminals
    t1: /\\/;
    t2: /a/;
    """)
    t1 = grammar.get_terminal('t1')
    assert t1.recognizer._regex == '\\\\'
    assert t1.recognizer('\\', 0) == '\\'


def test_builtin_grammar_action():
    """
    Builtin actions can be referenced from a grammar.
    """

    grammar = """
    @collect
    Ones: Ones One | One;

    terminals
    One: "1";
    """

    g = Grammar.from_string(grammar)

    ones = g.get_nonterminal('Ones')
    from parglare.actions import collect
    assert ones.action == collect

    p = Parser(g)
    result = p.parse('1 1 1 1 1')

    assert result == "1 1 1 1 1".split()


def test_multiple_grammar_action_raises_error():
    """
    If multiple actions are given for the same non-terminal GrammarError
    should be raised.
    """

    grammar = """
    S: Ones;

    @collect
    Ones: Ones One | One;

    @something
    Ones: 'foo';

    terminals
    One: "1";
    """

    # Actions 'collect' and 'something' defined for rule 'Ones'
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'Multiple' in str(e)


def test_action_override():
    """
    Explicitely provided action in `actions` param overrides default or
    grammar provided.
    """
    grammar = """
    S: Foo Bar;
    @pass_nochange
    Bar: "1" a;

    terminals
    @pass_nochange
    Foo: 'foo';
    a: "a";
    """

    g = Grammar.from_string(grammar)
    p = Parser(g)
    input_str = "foo 1 a"
    result = p.parse(input_str)
    assert result == ["foo", ["1", "a"]]

    actions = {
        "Foo": lambda _, __: "eggs",
        "Bar": lambda _, __: "bar reduce"}

    p = Parser(g, actions=actions)
    result = p.parse(input_str)
    assert result == ["eggs", "bar reduce"]

    # Test with actions call postponing
    p = Parser(g, build_tree=True, actions=actions)
    tree = p.parse(input_str)
    result = p.call_actions(tree)
    assert result == ["eggs", "bar reduce"]


def assignment_in_productions(prods, symbol_name, assgn_name):
    found = False
    for p in prods:
        if p.symbol.name == symbol_name:
            found = assgn_name in p.assignments
    return found


def test_assignment_plain():
    """
    Test plain assignment.
    """

    grammar = """
    S: "1" first=some_match "3";

    terminals
    some_match: "2";
    """

    g = Grammar.from_string(grammar)
    assert assignment_in_productions(g.productions, 'S', 'first')

    called = [False]

    def act_s(_, nodes, first):
        called[0] = True
        assert first == "2"
        return nodes

    actions = {
        "S": act_s
    }

    p = Parser(g, actions=actions)

    input_str = '1 2 3'

    result = p.parse(input_str)
    assert result == ["1", "2", "3"]
    assert called[0]


def test_assignment_bool():
    """
    Test bool assignment.
    """

    grammar = """
    S: "1" first?=some_match "3";

    terminals
    some_match: "2";
    """

    g = Grammar.from_string(grammar)
    assert assignment_in_productions(g.productions, 'S', 'first')

    called = [False]

    def act_s(_, nodes, first):
        called[0] = True
        assert first is True
        return nodes

    actions = {
        "S": act_s
    }

    p = Parser(g, actions=actions)

    input_str = '1 2 3'

    result = p.parse(input_str)
    assert result == ["1", "2", "3"]
    assert called[0]


def test_assignment_of_repetition():
    """
    Test assignment of repetition.
    """

    grammar = """
    S: "1" first=some_match+ "3";

    terminals
    some_match: "2";
    """

    g = Grammar.from_string(grammar)
    assert assignment_in_productions(g.productions, 'S', 'first')

    called = [False]

    def act_s(_, nodes, first):
        called[0] = True
        assert first == ["2", "2"]
        return nodes

    actions = {
        "S": act_s
    }

    p = Parser(g, actions=actions)

    input_str = '1 2 2 3'

    result = p.parse(input_str)
    assert result == ["1", ["2", "2"], "3"]
    assert called[0]


def test_assignment_of_repetition_with_sep():
    """
    Test assignment of repetition.
    """

    grammar = """
    S: "1" first=some_match+[comma] "3";

    terminals
    some_match: "2";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert assignment_in_productions(g.productions, 'S', 'first')

    called = [False]

    def act_s(_, nodes, first):
        called[0] = True
        assert first == ["2", "2"]
        return nodes

    actions = {
        "S": act_s
    }

    p = Parser(g, actions=actions)

    input_str = '1 2, 2 3'

    result = p.parse(input_str)
    assert result == ["1", ["2", "2"], "3"]
    assert called[0]


def test_multiple_assignment_with_repetitions():
    """
    Test assignment of repetition.
    """

    grammar = """
    S: "1" first=some_match+[comma] second?=some_match* "3";

    terminals
    some_match: "2";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert assignment_in_productions(g.productions, 'S', 'first')
    assert assignment_in_productions(g.productions, 'S', 'second')

    called = [False]

    def act_s(_, nodes, first, second):
        called[0] = True
        assert first == ["2", "2"]
        assert second is True
        return nodes

    actions = {
        "S": act_s
    }

    p = Parser(g, actions=actions)

    input_str = '1 2, 2 2 2 2 3'

    result = p.parse(input_str)
    assert result == ["1", ["2", "2"], ["2", "2", "2"], "3"]
    assert called[0]


def test_case_insensitive_parsing():
    """
    By default parglare is case sensitive. This test parsing without case
    sensitivity.
    """

    grammar = r"""
    S: "one" "Two" Astart;

    terminals
    Astart: /Aa\w+/;
    """

    g = Grammar.from_string(grammar)

    # By default parsing is case sensitive for both string and regex matches.
    parser = Parser(g)
    with pytest.raises(ParseError):
        parser.parse('One Two Aaa')

    with pytest.raises(ParseError):
        parser.parse('one Two AAa')

    g = Grammar.from_string(grammar, ignore_case=True)
    parser = Parser(g)
    parser.parse('One Two Aaa')
    parser.parse('one Two AAa')



from parglare import Grammar, GLRParser


def test_regex_alternative_match_bug():
    """
    """

    grammar = """
    A: "Begin" Eq "End";

    terminals
    Eq: /=|EQ/;
    """
    g = Grammar.from_string(grammar)
    parser = GLRParser(g)
    parser.parse('Begin EQ End')



import pytest
import os
from types import FunctionType
from parglare import Grammar, Parser, GrammarError, ParseError
from .base_recognizers import number

this_folder = os.path.dirname(__file__)

model_str = '''
modelID 42.23.5
component myComponent extends some.fqn.name {
    in SomeInputSlot
    out SomeOutputSlot
}
'''


def comma_recognizer(input, pos):
    if input[pos] == ',':
        return input[pos:pos + 1]


def test_imported_recognizers_error_undefined_recognizer():

    with pytest.raises(GrammarError,
                       match=r'has no recognizer defined and no recognizers '
                       'are given'):
        Grammar.from_file(os.path.join(this_folder, 'model.pg'))

    # If we define COMMA recognizer grammar will construct without exceptions.
    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'),
                          recognizers={'base.COMMA': comma_recognizer})
    assert g


def test_imported_recognizers_connect_from_external_file():
    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'),
                          recognizers={'base.COMMA': comma_recognizer})

    # Check that recognizers are loaded and connected.
    rec_fqn = g.get_terminal('base.FQN')
    assert rec_fqn.recognizer
    assert type(rec_fqn.recognizer) is FunctionType
    assert rec_fqn.recognizer.__name__ == 'FQN'

    rec_fqn = g.get_terminal('base.NUMERIC_ID')
    assert rec_fqn.recognizer
    assert type(rec_fqn.recognizer) is FunctionType
    assert rec_fqn.recognizer.__name__ == 'number'


def test_imported_recognizers_override():
    """
    Test that recognizers loaded from `*_recognizers.py` files can be
    overriden by users provided recognizers.
    """

    called = [False, False]

    def numeric_id(input, pos):
        called[0] = True

    def fqn(input, pos):
        called[0] = True

    recognizers = {
        'base.COMMA': comma_recognizer,
        'base.NUMERIC_ID': numeric_id,
        'base.FQN': fqn
    }

    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'),
                          recognizers=recognizers)
    assert g
    with pytest.raises(ParseError):
        Parser(g).parse(model_str)
    assert any(called)

    called = [False]

    def numeric_id(input, pos):
        called[0] = True
        return number(input, pos)

    recognizers = {
        'base.COMMA': comma_recognizer,
        'base.NUMERIC_ID': numeric_id,
    }

    g = Grammar.from_file(os.path.join(this_folder, 'model.pg'),
                          recognizers=recognizers)
    assert g
    Parser(g).parse(model_str)
    assert called[0]


def test_imported_recognizers_override_by_importing_grammar_file():
    """
    Test that recognizers loaded from `*_recognizers.py` files can be
    overriden in importing grammar `*_recognizers.py` file by providing
    FQN of the imported terminal relative from the importing grammar file.
    """

    g = Grammar.from_file(os.path.join(this_folder, 'model_override.pg'))
    assert g

    t = g.get_terminal('base.NUMERIC_ID')
    assert t is not None

    assert t.recognizer.__doc__ == 'Check override'



import os
from parglare import Parser, Grammar


def test_table_from_cache_different_than_calculated():
    this_folder = os.path.dirname(__file__)
    grammar_file = os.path.join(this_folder, 'grammar.pg')
    table_file = os.path.join(this_folder, 'grammar.pgt')
    try:
        os.remove(table_file)
    except Exception:
        pass

    g = Grammar.from_file(grammar_file)

    p = Parser(g)
    without_cache = p.parse('dynamic("OS")')

    p = Parser(g)
    with_cache = p.parse('dynamic("OS")')

    assert without_cache == with_cache



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
from parglare import Grammar, Parser
import re


def test_recognizer_context():
    grammar = """
    program: expression+[semicolon];
    expression: term+;

    terminals
    semicolon: ";";
    term:;
    """

    term_re = re.compile(r"[a-zA-Z_]+")

    def term(context, input, pos):
        match = term_re.match(input, pos)
        if match is None:
            return None
        return input[pos:match.end()]

    g = Grammar.from_string(grammar, recognizers={'term': term})
    parser = Parser(g)
    assert parser.parse("a bb cc; d ee f; g hh i")



import os
from parglare import Grammar

this_folder = os.path.dirname(__file__)


def test_fqn_constructed_by_first_import_path():

    g = Grammar.from_file(os.path.join(this_folder, 'A.pg'))

    assert g.get_terminal('B.C.CTerm')
    assert not g.get_terminal('C.CTerm')
    assert g.get_nonterminal('B.C.CRule')
    assert not g.get_nonterminal('C.CRule')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest
from parglare import Parser, Grammar
from parglare.exceptions import GrammarError, ParseError, RRConflicts


def test_repeatable_zero_or_more():
    """
    Tests zero or more repeatable operator.
    """

    grammar = """
    S: "2" b* "3";

    terminals
    b: "1";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_0')
    assert g.get_nonterminal('b_1')

    p = Parser(g)

    input_str = '2 1 1 1 3'
    result = p.parse(input_str)
    assert result == ["2", ["1", "1", "1"], "3"]

    input_str = '2 3'
    result = p.parse(input_str)
    assert result == ["2", [], "3"]


def test_repeatable_zero_or_more_with_separator():
    """
    Tests zero or more repeatable operator with separator.
    """

    grammar = """
    S: "2" b*[comma] "3";

    terminals
    b: "1";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_0_comma')

    p = Parser(g)

    input_str = '2 1, 1 , 1 3'
    result = p.parse(input_str)
    assert result == ["2", ["1", "1", "1"], "3"]

    input_str = '2 3'
    result = p.parse(input_str)
    assert result == ["2", [], "3"]


def test_repeatable_one_or_more():
    """
    Tests one or more repeatable operator.
    """

    grammar = """
    S: "2" b+ "3";

    terminals
    b: "1";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_1')

    p = Parser(g)

    input_str = '2 1 1 1 3'
    result = p.parse(input_str)
    assert result == ["2", ["1", "1", "1"], "3"]

    input_str = '2 3'
    with pytest.raises(ParseError) as e:
        result = p.parse(input_str)
    assert 'Expected: b' in str(e)


def test_repeatable_one_or_more_with_separator():
    """
    Tests one or more repeatable operator with separator.
    """

    grammar = """
    S: "2" b+[comma] "3";

    terminals
    b: "1";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_1_comma')

    p = Parser(g)

    input_str = '2 1, 1 , 1 3'
    result = p.parse(input_str)
    assert result == ["2", ["1", "1", "1"], "3"]

    input_str = '2 3'
    with pytest.raises(ParseError) as e:
        p.parse(input_str)
    assert 'Expected: b' in str(e)


def test_optional():
    """
    Tests optional operator.
    """

    grammar = """
    S: "2" b? "3"? EOF;

    terminals
    b: "1";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_opt')

    p = Parser(g)

    input_str = '2 1 3'
    result = p.parse(input_str)
    assert result == ["2", "1", "3", None]

    input_str = '2 3'
    result = p.parse(input_str)
    assert result == ["2", None, "3", None]

    input_str = '2 1'
    result = p.parse(input_str)
    assert result == ["2", "1", None, None]

    input_str = ' 1 3'
    with pytest.raises(ParseError) as e:
        p.parse(input_str)
    assert 'Expected: 2' in str(e)


def test_optional_no_modifiers():
    """
    Tests that optional operator doesn't allow modifiers.
    """

    grammar = """
    S: "2" b?[comma] "3"? EOF;

    terminals
    b: "1";
    comma: ",";
    """

    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert "Repetition modifier not allowed" in str(e)


def test_multiple_repetition_operators():
    """
    Test using of multiple repetition operators.
    """
    grammar = """
    S: "2" b*[comma] c+ "3"? EOF;

    terminals
    b: "b";
    c: "c";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_0_comma')
    assert g.get_nonterminal('c_1')

    p = Parser(g)

    input_str = '2 b, b  c 3'
    result = p.parse(input_str)
    assert result == ["2", ["b", "b"], ["c"], "3", None]


def test_repetition_operator_many_times_same():
    """
    Test using the same repetition operator multiple times.
    """

    grammar = """
    S: "2" b*[comma] "3"? b*[comma] EOF;

    terminals
    b: "b";
    comma: ",";
    """

    g = Grammar.from_string(grammar)
    assert g.get_nonterminal('b_0_comma')

    p = Parser(g)

    input_str = '2 b  3 b, b'
    result = p.parse(input_str)
    assert result == ["2", ["b"], "3", ["b", "b"], None]


def test_repeatable_one_zero_rr_conflicts():
    """
    Check that translations of B+ and B* don't produce R/R conflict.
    """
    grammar = """
    S: A B+ C;
    S: A B* D;

    terminals
    A:; B:; C:; D:;
    """
    g = Grammar.from_string(grammar, _no_check_recognizers=True)

    # Check if parser construction raises exception
    try:
        Parser(g)
    except RRConflicts:
        pytest.fail("R/R conflicts not expected here.")



from parglare import Grammar, Parser, get_collector

grammar_1 = r'''
STMT : STMT ADDOP STMT {left, 1}
     | STMT MULOP STMT {left, 1}
     | "(" STMT ")" | NUMBER;
ADDOP : "+" {1}
      | "-" {1};
MULOP : "*" {2}
      | "/" {2};

terminals
NUMBER: /\d+(.\d+)?/;
'''

grammar_2 = r'''
STMT : STMT "+" STMT {left, 1}
     | STMT "-" STMT {left, 1}
     | STMT "*" STMT {left, 2}
     | STMT "/" STMT {left, 2}
     | "(" STMT ")" | NUMBER;

terminals
NUMBER: /\d+(.\d+)?/;
'''

# Grammar could be simplified a bit
# with https://github.com/igordejanovic/parglare/issues/17
grammar_3 = r'''
STMT {left}: STMT ADDOP STMT {1}
           | STMT MULOP STMT {2};
STMT: "(" STMT ")" | NUMBER;
ADDOP {1}: "+" | "-";
MULOP {2}: "*" | "/";

terminals
NUMBER: /\d+(.\d+)?/;
'''

expression = '1 - 2 / (3 - 1 + 5 / 6 - 8 + 8 * 2 - 5)'
result = 1 - 2 / (3 - 1 + 5 / 6 - 8 + 8 * 2 - 5)

action = get_collector()


@action
def NUMBER(_, value):
    return int(value)


@action
def STMT(_, nodes):
    if len(nodes) == 1:
        return nodes[0]
    elif nodes[0] == '(':
        return nodes[1]
    else:
        left, op, right = nodes

    if op == '+':
        return left + right
    elif op == '-':
        return left - right
    elif op == '*':
        return left * right
    else:
        return left / right


def test_associativity_variant_1():
    """
    See https://github.com/igordejanovic/parglare/issues/22
    """
    g = Grammar.from_string(grammar_1)
    parser = Parser(g, actions=action.all)
    r = parser.parse(expression)

    assert r == result


def test_associativity_variant_2():
    """
    See https://github.com/igordejanovic/parglare/issues/22
    """
    g = Grammar.from_string(grammar_2)
    parser = Parser(g, actions=action.all)
    r = parser.parse(expression)

    assert r == result


def test_associativity_variant_3():
    """
    See https://github.com/igordejanovic/parglare/issues/22
    Using https://github.com/igordejanovic/parglare/issues/17
    """
    g = Grammar.from_string(grammar_3)
    parser = Parser(g, actions=action.all)
    r = parser.parse(expression)

    assert r == result



from parglare.grammar import EMPTY, NonTerminal

LR_0 = 0
LR_1 = 1


def closure(state, itemset_type, first_sets=None):
    """
    For the given LRState calculates its LR(0)/LR(1) itemset closure.

    Args:
    state(LRState):
    itemset_type(int): LR_0 or LR_1
    first_sets(dict of sets): Used in LR_1 itemsets calculation.
    """
    from parglare.tables import LRItem

    while True:

        has_additions = False
        to_add = []
        for item in state.items:
            gs = item.symbol_at_position
            if isinstance(gs, NonTerminal):
                for p in state.grammar.productions:

                    if p.symbol == gs:

                        if itemset_type is LR_1:
                            # Calculate follow set that is possible after the
                            # non-terminal at the given position of the current
                            # item.
                            follow = _new_item_follow(item, first_sets)
                            new_item = LRItem(p, 0, follow)
                        else:
                            new_item = LRItem(p, 0)

                        if new_item not in state.items \
                                and new_item not in to_add:

                            # If the item doesn't exists yet add it.
                            to_add.append(new_item)
                            has_additions = True

                        else:
                            # If the item already exists, this newly created
                            # item might still have a wider follows set. If so,
                            # update with the current new item follows set if
                            # we are building LR_1 items set.
                            if itemset_type is LR_1:
                                try:
                                    existing_item = state.items[
                                        state.items.index(new_item)]
                                except ValueError:
                                    existing_item = to_add[
                                        to_add.index(new_item)]

                                if not follow.issubset(existing_item.follow):
                                    existing_item.follow.update(follow)
                                    has_additions = True

        if has_additions:
            state.items.extend(to_add)
        else:
            break


def _new_item_follow(item, first_sets):
    """
    Returns follow set of possible terminals after the item's current
    non-terminal.

    Args:
    item (LRItem): The source item which is causing the creation of the
        new item.
    first_sets(dict of sets): The dict of set of first items keyed by
        a grammar symbol.
    """

    new_follow = set()
    for s in item.production.rhs[item.position + 1:]:
        new_follow.update(first_sets[s])
        if EMPTY not in new_follow:
            # If EMPTY can't be derived at current position than we have found
            # the whole follow set.
            break
        else:
            # If the EMPTY is possible at current position in this loop we must
            # continue to include firsts of the next grammar symbol.
            # EMTPY can't be a member of the follow set.
            new_follow.remove(EMPTY)
    else:
        # If the rest of production can be EMPTY we shall inherit
        # all elements of the source item follow set.
        new_follow.update(item.follow)

    return new_follow



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
from parglare import Grammar, Parser, ParseError, ParserInitError, \
    GrammarError, DisambiguationError
from parglare.actions import pass_single, pass_nochange, collect


def test_parse_list_of_integers():

    grammar = """
    Numbers: all_less_than_five EOF;
    all_less_than_five: all_less_than_five int_less_than_five
                      | int_less_than_five;

    terminals
    int_less_than_five:;
    """

    def int_less_than_five(input, pos):
        if input[pos] < 5:
            return [input[pos]]

    recognizers = {
        'int_less_than_five': int_less_than_five
    }
    g = Grammar.from_string(grammar, recognizers=recognizers, debug=True)

    actions = {
        'Numbers': pass_single,
        'all_less_than_five': collect,
        'int_less_than_five': pass_single
    }

    # Test that `ws` must be set to `None` for non-textual content
    parser = Parser(g, actions=actions)

    ints = [3, 4, 1, 4]
    with pytest.raises(
            ParserInitError,
            match=r'For parsing non-textual content please '
            'set `ws` to `None`'):
        parser.parse(ints)

    parser = Parser(g, actions=actions, ws=None)
    ints = [3, 4, 1, 4]
    p = parser.parse(ints)
    assert p == ints

    # Test that error is correctly reported.
    with pytest.raises(ParseError) as e:
        parser.parse([4, 2, 1, 6, 3])
    assert '1:3:"[4, 2, 1]*[6, 3]"' in str(e)
    assert 'int_less_than_five' in str(e)


def test_parse_list_of_integers_lexical_disambiguation():

    def int_less_than_five(input, pos):
        if input[pos] < 5:
            return [input[pos]]

    def ascending(input, pos):
        "Match sublist of ascending elements. Matches at least one."
        last = pos + 1
        while last < len(input) and input[last] > input[last-1]:
            last += 1
        if last > pos:
            return input[pos:last]

    def ascending_nosingle(input, pos):
        "Match sublist of ascending elements. Matches at least two."
        last = pos + 1
        while last < len(input) and input[last] > input[last-1]:
            last += 1
        if last - pos >= 2:
            return input[pos:last]

    grammar = """
    Numbers: all_less_than_five ascending all_less_than_five EOF;
    all_less_than_five: all_less_than_five int_less_than_five
                      | int_less_than_five;

    terminals
    int_less_than_five:;
    ascending:;
    """

    recognizers = {
        'int_less_than_five': int_less_than_five,
        'ascending': ascending
    }
    g = Grammar.from_string(grammar, recognizers=recognizers)

    actions = {
        'Numbers': lambda _, nodes: [nodes[0], nodes[1], nodes[2]],
        'all_less_than_five': collect,
        'int_less_than_five': pass_single,   # Unpack element for collect
        'ascending': pass_nochange
    }
    parser = Parser(g, actions=actions, ws=None, debug=True)

    ints = [3, 4, 1, 4, 7, 8, 9, 3]

    # This must fail as ascending and int_less_than_five recognizers both
    # might match just a single int and after parser has saw 3 it will try
    # to disambiguate and fail as the following 4 is recognized by both
    # recognizers.
    with pytest.raises(DisambiguationError):
        p = parser.parse(ints)

    # Now we change the recognizer for ascending to match at least two
    # consecutive ascending numbers.
    recognizers['ascending'] = ascending_nosingle
    g = Grammar.from_string(grammar, recognizers=recognizers)
    parser = Parser(g, actions=actions, ws=None, debug=True)

    # Parsing now must pass
    p = parser.parse(ints)

    assert p == [[3, 4], [1, 4, 7, 8, 9], [3]]


def test_terminals_with_emtpy_bodies_require_recognizers():
    """
    If there are terminals with empty bodies in the grammar then recognizers
    must be given and there must be a recognizer for each terminal missing
    in-grammar recognizer.
    """

    grammar = """
    S: A | B | C;

    terminals
    A: {15};
    B: ;
    C: "c";
    """

    with pytest.raises(GrammarError):
        g = Grammar.from_string(grammar)

    recognizers = {
        'B': lambda input, pos: None,
    }

    with pytest.raises(GrammarError):
        g = Grammar.from_string(grammar, recognizers=recognizers)

    recognizers['A'] = lambda input, pos: None

    g = Grammar.from_string(grammar, recognizers=recognizers)
    assert g

    # Test that setting _no_check_recognizers will prevent grammar
    # error. This is used in pglr command.
    Grammar.from_string(grammar, _no_check_recognizers=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pytest  # noqa
import os
from parglare import Grammar, Parser, GLRParser, ParseError
from .expression_grammar import get_grammar


parsers = pytest.mark.parametrize("parser_class", [Parser, GLRParser])


@parsers
def test_invalid_input(parser_class):

    grammar = get_grammar()
    p = parser_class(grammar)

    with pytest.raises(ParseError) as e:
        p.parse("id+id*+id")

    assert e.value.location.start_position == 6
    assert "(" in str(e)
    assert "id" in str(e)
    assert '*' in [s.name for s in e.value.symbols_before]
    assert '+' in [t.value for t in e.value.tokens_ahead]
    expected_names = [s.name for s in e.value.symbols_expected]
    assert 'id' in expected_names
    assert '(' in expected_names


@parsers
def test_premature_end(parser_class):

    grammar = get_grammar()
    p = parser_class(grammar)

    with pytest.raises(ParseError) as e:
        p.parse("id+id*")

    assert e.value.location.start_position == 6
    expected_names = [s.name for s in e.value.symbols_expected]
    assert 'id' in expected_names
    assert '(' in expected_names
    assert '*' in [s.name for s in e.value.symbols_before]
    assert e.value.tokens_ahead == []


def test_ambiguous_glr():
    grammar = r"""
    S: E EOF;
    E: E '+' E
     | E '*' E
     | number;

    terminals
    number: /\d+(\.\d+)?/;
    """
    g = Grammar.from_string(grammar)
    parser = GLRParser(g)

    with pytest.raises(ParseError) as e:
        parser.parse("1 + 2 * 3 / 5")

    assert e.value.location.start_position == 10
    assert 'number' in [s.name for s in e.value.symbols_before]


@parsers
def test_line_column(parser_class):
    grammar = get_grammar()
    p = parser_class(grammar)

    with pytest.raises(ParseError) as e:
        p.parse("""id + id * id + id + error * id""")

    loc = e.value.location
    assert loc.start_position == 20
    assert loc.line == 1
    assert loc.column == 20

    with pytest.raises(ParseError) as e:
        p.parse("""id + id * id + id + error * id

        """)
    loc = e.value.location
    assert loc.start_position == 20
    assert loc.line == 1
    assert loc.column == 20

    with pytest.raises(ParseError) as e:
        p.parse("""

id + id * id + id + error * id""")
    loc = e.value.location
    assert loc.start_position == 22
    assert loc.line == 3
    assert loc.column == 20

    with pytest.raises(ParseError) as e:
        p.parse("""

id + id * id + id + error * id

        """)
    loc = e.value.location
    assert loc.start_position == 22
    assert loc.line == 3
    assert loc.column == 20


@parsers
def test_file_name(parser_class):
    "Test that file name is given in the error string when parsing file."
    grammar = get_grammar()
    p = parser_class(grammar)

    input_file = os.path.join(os.path.dirname(__file__),
                              'parsing_errors.txt')

    with pytest.raises(ParseError) as e:
        p.parse_file(input_file)

    assert 'parsing_errors.txt' in str(e.value)
    assert 'parsing_errors.txt' in e.value.location.file_name



import os
import filecmp
from parglare import Grammar, Parser

this_folder = os.path.dirname(__file__)

input_str = '''

package First
package Second {
    component packageComponent {

    }
}

module SomeModule {

    component myComponent {
        in SomeInputSlot
        out SomeOutputSlot
    }


}

'''


def test_diamond_import_resolving_and_model_creation():

    grammar_file = os.path.join(this_folder, 'model.pg')
    table_file = os.path.join(this_folder, 'model.pgt')
    table_cmp_file = os.path.join(this_folder, 'model_compare.pgt')

    g = Grammar.from_file(grammar_file)
    try:
        os.remove(table_file)
    except Exception:
        pass

    parser = Parser(g)

    # Check generated table file.
    assert filecmp.cmp(table_file, table_cmp_file, shallow=False)

    # Check that parser loaded from the table will correctly parse
    parser = Parser(g, force_load_table=True)

    model = parser.parse(input_str)
    assert model
    assert model.__class__.__name__ == 'Model'
    assert type(model.packages) is list
    assert len(model.packages) == 2
    assert model.packages[0].name == 'First'
    assert type(model.modules) is list
    assert len(model.modules) == 1

    packageComponent = model.packages[1].body.components[0]
    assert packageComponent.name == 'packageComponent'

    module = model.modules[0]
    assert module.__class__.__name__ == 'm.Module'
    assert module.name == 'SomeModule'
    assert len(module.components) == 1

    component = module.components[0]
    assert type(component) == type(packageComponent)
    assert component.name == 'myComponent'
    assert len(component.slots) == 2

    slot = component.slots[1]
    assert slot.__class__.__name__ == 'packages.components.SlotOut'
    assert slot.name == 'SomeOutputSlot'



def act_assignment(context, nodes):
    name = nodes[0]
    number = nodes[2]

    # Use context.extra to collect variables
    if context.extra is None:
        context.extra = {}

    context.extra[name] = number


actions = {
    "Calc": lambda _, nodes: nodes[1],
    "Assignment": act_assignment,
    "E": [lambda _, nodes: nodes[0] + nodes[2],
          lambda _, nodes: nodes[0] - nodes[2],
          lambda _, nodes: nodes[0] * nodes[2],
          lambda _, nodes: nodes[0] / nodes[2],
          lambda _, nodes: nodes[1],
          lambda _, nodes: nodes[0],
          lambda _, nodes: nodes[0]],
    "Number": lambda _, value: float(value),
    "VariableName": lambda _, value: value,
    "VariableRef": lambda context, nodes: context.extra[nodes[0]],
}



from __future__ import unicode_literals
from parglare import get_collector

recognizer = get_collector()


@recognizer('base.NUMERIC_ID')
def number(input, pos):
    '''Check override'''
    pass


@recognizer('base.COMMA')
def comma_recognizer(input, pos):
    if input[pos] == ',':
        return input[pos:pos + 1]
